# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from collections.abc import Callable
from typing import Dict, Set

from xpra.os_util import gi_import
from xpra.util.str_fn import strtobytes, csv

from xpra.log import Logger
log = Logger("x11", "bindings", "gtk")

GObject = gi_import("GObject")
GdkX11= gi_import("GdkX11")
Gdk = gi_import("Gdk")
Gtk = gi_import("Gtk")


from xpra.x11.bindings.xlib cimport (
    Display, Window, Visual, Atom,
    XEvent, XGetErrorText, XGetAtomName, XFree,
)
from xpra.x11.bindings.events cimport parse_xevent, init_x11_events
from xpra.x11.bindings.events import get_x_event_signals, get_x_event_type_name

from libc.stdint cimport uintptr_t
from xpra.gtk.bindings.gobject cimport wrap, unwrap


from xpra.x11.common import REPR_FUNCTIONS
def get_window_xid(window) -> str:
    return hex(window.get_xid())
REPR_FUNCTIONS[GdkX11.X11Window] = get_window_xid
def get_display_name(display) -> str:
    return display.get_name()
REPR_FUNCTIONS[Gdk.Display] = get_display_name


cdef extern from "gdk_x11_macros.h":
    int is_x11_display(void *)


cdef GdkDisplay * get_raw_display(display) except? NULL:
    return <GdkDisplay*> unwrap(display, Gdk.Display)


def is_X11_Display(display=None) -> bool:
    cdef GdkDisplay *gdk_display
    if display is None:
        manager = Gdk.DisplayManager.get()
        display = manager.get_default_display()
    if display is None:
        log("no default display!")
        return False
    try:
        gdk_display = get_raw_display(display)
    except TypeError:
        return False
    return bool(is_x11_display(gdk_display))

###################################
# Headers, python magic
###################################
cdef extern from "gtk-3.0/gdk/gdk.h":
    ctypedef void* GdkAtom  # @UndefinedVariable
    GdkAtom GDK_NONE
    ctypedef struct GdkWindow:
        pass
    void gdk_display_flush(GdkDisplay *display)
    void gdk_x11_display_error_trap_push(GdkDisplay *display)
    int gdk_x11_display_error_trap_pop(GdkDisplay *display)

cdef extern from "gtk-3.0/gdk/gdkx.h":
    pass

cdef extern from "gtk-3.0/gdk/gdkproperty.h":
    ctypedef int gint
    ctypedef gint gboolean
    char* gdk_atom_name(GdkAtom atom)
    GdkAtom gdk_atom_intern(const char *atom_name, gboolean only_if_exists)

cdef extern from "glib-2.0/glib-object.h":
    ctypedef struct cGObject "GObject":
        pass

cdef extern from "pygobject-3.0/pygobject.h":
    cGObject *pygobject_get(object box)
    object pygobject_new(cGObject * contents)

    ctypedef void* gpointer  # @UndefinedVariable
    ctypedef int GType
    ctypedef struct PyGBoxed:
        #PyObject_HEAD
        gpointer boxed
        GType gtype

######
# GDK primitives, and wrappers for Xlib
######

# gdk_region_get_rectangles (pygtk bug #517099)
cdef extern from "gtk-3.0/gdk/gdktypes.h":
    ctypedef struct cGdkVisual "GdkVisual":
        pass
    Visual * GDK_VISUAL_XVISUAL(cGdkVisual   *visual)

    Window GDK_WINDOW_XID(GdkWindow *)

    ctypedef struct GdkDisplay:
        pass
    Display * GDK_DISPLAY_XDISPLAY(GdkDisplay *)

    GdkDisplay * gdk_x11_lookup_xdisplay(Display *)

    # FIXME: this should have stricter type checking
    object PyGdkAtom_New(GdkAtom)
    Atom gdk_x11_get_xatom_by_name(char *atom_name)


# Basic utilities:

def get_pywindow(Window xwindow) -> GdkX11.X11Window | None:
    if xwindow==0:
        return None
    display = Gdk.get_default_root_window().get_display()
    try:
        return GdkX11.X11Window.foreign_new_for_display(display, xwindow)
    except TypeError as e:
        log("cannot get gdk window for %s : %#x, %s", display, xwindow, e)
    return None


cdef Display * get_xdisplay() except? NULL:
    gdk_display = Gdk.get_default_root_window().get_display()
    return GDK_DISPLAY_XDISPLAY(get_raw_display(gdk_display))


def get_xatom(str_or_xatom) -> Atom:
    """Returns the X atom corresponding to the given Python string or Python
    integer (assumed to already be an X atom)."""
    if isinstance(str_or_xatom, int):
        i = int(str_or_xatom)
        assert i>=0, "invalid int atom value %s" % str_or_xatom
        return i
    assert isinstance(str_or_xatom, str), "argument is not a string or number: %s" % type(str_or_xatom)
    gdkatom = Gdk.atom_intern(str_or_xatom, False)
    if not gdkatom:
        return  0
    #atom_intern takes str, but gdk_x11_get_xatom_by_name wants bytes!
    b = strtobytes(str_or_xatom)
    return gdk_x11_get_xatom_by_name(b)

def get_pyatom(xatom) -> str:
    return _get_pyatom(xatom)

cdef object _get_pyatom(int xatom):
    if int(xatom) > 2**32:
        raise ValueError(f"weirdly huge purported xatom: {xatom}")
    if xatom==0:
        return ""
    display = Gdk.get_default_root_window().get_display()
    cdef GdkDisplay *disp = get_raw_display(display)
    cdef Display *xdisplay = GDK_DISPLAY_XDISPLAY(disp)
    cdef char *name = XGetAtomName(xdisplay, xatom)
    if name==NULL:
        return ""
    pyname = name.decode("latin1")
    XFree(name)
    return pyname


###################################
# Event handling
###################################

# We need custom event handling in a few ways:
#   -- We need to listen to various events on client windows, even though they
#      have no GtkWidget associated.
#   -- We need to listen to various events that are not otherwise wrapped by
#      GDK at all.  (In particular, the SubstructureRedirect events.)
# To do this, we use two different hooks in GDK:
#   gdk_window_add_filter: This allows us to snoop on all events before they
#     are converted into GDK events.
# Our hooks in any case use the "xpra-route-events-to" GObject user data
# field of the gdk.Window's involved.  For the SubstructureRedirect
# events, we use this field of either the window that is making the request,
# or, if its field is unset, to the window that actually has
# SubstructureRedirect selected on it; for other events, we send it to the
# event window directly.
#
# So basically, to use this code:
#   -- Import this module to install the global event filters
#   -- Call win.set_data("xpra-route-events-to", obj) on random windows.
#   -- Call addXSelectInput or its convenience wrappers, substructureRedirect
#      and selectFocusChange.
#   -- Receive interesting signals on 'obj'.


cdef extern from "gtk-3.0/gdk/gdkevents.h":
    ctypedef enum GdkFilterReturn:
        GDK_FILTER_CONTINUE   # If we ignore the event
        GDK_FILTER_TRANSLATE  # If we converted the event to a GdkEvent
        GDK_FILTER_REMOVE     # If we handled the event and GDK should ignore it

    ctypedef struct GdkXEvent:
        pass
    ctypedef struct GdkEvent:
        pass

    ctypedef GdkFilterReturn (*GdkFilterFunc)(GdkXEvent *, GdkEvent *, void *) except GDK_FILTER_CONTINUE
    void gdk_window_add_filter(GdkWindow * w, GdkFilterFunc function, void * userdata)

    void gdk_window_remove_filter(GdkWindow *window, GdkFilterFunc function, void * data)


# No need to select for ClientMessage; in fact, one cannot select for
# ClientMessages.  If they are sent with an empty mask, then they go to the
# client that owns the window they are sent to, otherwise they go to any
# clients that are selecting for that mask they are sent with.

event_receivers_map : Dict[int,Set[Callable]] = {}

def add_event_receiver(xid: int, receiver:callable, max_receivers: int=3) -> None:
    if not isinstance(xid, int):
        raise TypeError(f"xid must be an int, not a {type(xid)}")
    receivers = event_receivers_map.get(xid)
    if receivers is None:
        receivers = set()
        event_receivers_map[xid] = receivers
    if max_receivers>0 and len(receivers)>max_receivers:
        from xpra.x11.window_info import window_info
        log.warn("Warning: already too many event receivers")
        log.warn(f" for {window_info(xid)!r}: {len(receivers)}")
        log.warn(f" adding {receiver!r} to {receivers}", backtrace=True)
    receivers.add(receiver)

def remove_event_receiver(xid: int, receiver:callable) -> None:
    if not isinstance(xid, int):
        raise TypeError(f"xid must be an int, not a {type(xid)}")
    receivers = event_receivers_map.get(xid)
    if receivers is None:
        return
    receivers.discard(receiver)
    if not receivers:
        event_receivers_map.pop(xid)

#only used for debugging:
def get_event_receivers(xid: int) -> Set[Callable]:
    return event_receivers_map.get(xid)

def cleanup_all_event_receivers() -> None:
    event_receivers_map.clear()


#sometimes we may want to debug routing for certain X11 event types
debug_route_events : List[int] = []

def get_error_text(code) -> str:
    if type(code)!=int:
        return str(code)
    cdef Display *xdisplay = get_xdisplay()
    cdef char[128] buffer
    XGetErrorText(xdisplay, code, buffer, 128)
    return str(buffer[:128])


#and change this debugging on the fly, programmatically:
def add_debug_route_event(event_type: int) -> None:
    debug_route_events.append(event_type)
def remove_debug_route_event(event_type: int) -> None:
    debug_route_events.remove(event_type)


catchall_receivers : Dict[str,List[Callable]] = {}
def add_catchall_receiver(signal, handler) -> None:
    catchall_receivers.setdefault(signal, []).append(handler)
    log("add_catchall_receiver(%s, %s) -> %s", signal, handler, catchall_receivers)

def remove_catchall_receiver(signal, handler) -> None:
    receivers = catchall_receivers.get(signal)
    if receivers:
        receivers.remove(handler)
    log("remove_catchall_receiver(%s, %s) -> %s", signal, handler, catchall_receivers)


fallback_receivers : Dict[str,List[Callable]] = {}
def add_fallback_receiver(signal, handler) -> None:
    fallback_receivers.setdefault(signal, []).append(handler)
    log("add_fallback_receiver(%s, %s) -> %s", signal, handler, fallback_receivers)

def remove_fallback_receiver(signal, handler) -> None:
    receivers = fallback_receivers.get(signal)
    if receivers:
        receivers.remove(handler)
    log("remove_fallback_receiver(%s, %s) -> %s", signal, handler, fallback_receivers)


cdef void _maybe_send_event(unsigned int DEBUG, handlers, signal, event, hinfo="window"):
    if not handlers:
        if DEBUG:
            log.info("  no handler registered for %s (%s)", hinfo, handlers)
        return
    # Copy the 'handlers' list, because signal handlers might cause items
    # to be added or removed from it while we are iterating:
    for handler in tuple(handlers):
        signals = GObject.signal_list_names(handler)
        if signal in signals:
            if DEBUG:
                log.info("  forwarding event to a %s %s handler's %s signal", type(handler).__name__, hinfo, signal)
            handler.emit(signal, event)
            if DEBUG:
                log.info("  forwarded")
        elif DEBUG:
            log.info("  not forwarding to %s handler, it has no %r signal", type(handler).__name__, signal)
            log.info("     only: %s", csv(f"{s!r}" for s in signals))

cdef void _route_event(int etype, event, signal, parent_signal):
    # Sometimes we get GDK events with event.window == None, because they are
    # for windows we have never created a GdkWindow object for, and GDK
    # doesn't do so just for this event.  As far as I can tell this only
    # matters for override redirect windows when they disappear, and we don't
    # care about those anyway.
    cdef unsigned int DEBUG = etype in debug_route_events
    if DEBUG:
        log.info(f"{event}")
    handlers = None
    if event.window==event.delivered_to:
        window = event.window
        if signal is not None:
            if DEBUG:
                log.info(f"  delivering {signal!r} to window itself: {window:x}")
            if window:
                handlers = get_event_receivers(window)
                _maybe_send_event(DEBUG, handlers, signal, event, "window %#x" % window)
        elif DEBUG:
            log.info(f"  received event on window {window:x} itself but have no signal for that")
    else:
        window = event.delivered_to
        if parent_signal is not None:
            if not window:
                if DEBUG:
                    log.info(f"  event.delivered_to={window}, ignoring")
            else:
                if DEBUG:
                    log.info(f"  delivering {parent_signal!r} to parent window: {window:x}")
                handlers = get_event_receivers(window)
                _maybe_send_event(DEBUG, handlers, parent_signal, event, "parent window %#x" % window)
        else:
            if DEBUG:
                log.info(f"  received event on parent window {window:x} but have no parent signal")
    #fallback only fires if nothing else has fired yet:
    if not handlers:
        global fallback_receivers
        if signal:
            handlers = fallback_receivers.get(signal)
            _maybe_send_event(DEBUG, handlers, signal, event, "fallback-signal")
        if parent_signal:
            handlers = fallback_receivers.get(parent_signal)
            _maybe_send_event(DEBUG, handlers, parent_signal, event, "fallback-parent-signal")
    #always fire those:
    global catchall_receivers
    if signal:
        handlers = catchall_receivers.get(signal)
        _maybe_send_event(DEBUG, handlers, signal, event, "catchall-signal")
    if parent_signal:
        handlers = catchall_receivers.get(parent_signal)
        _maybe_send_event(DEBUG, handlers, parent_signal, event, "catchall-parent-signal")


cdef GdkFilterReturn x_event_filter(GdkXEvent * e_gdk,
                                    GdkEvent * gdk_event,
                                    void * userdata) except GDK_FILTER_CONTINUE with gil:
    cdef object event_args
    cdef object pyev
    cdef double start = monotonic()
    cdef int etype

    cdef Display *display = get_xdisplay()
    cdef XEvent * e = <XEvent*>e_gdk
    try:
        pyev = parse_xevent(display, e)
    except Exception:
        log.error("Error parsing X11 event", exc_info=True)
        return GDK_FILTER_CONTINUE  # @UndefinedVariable
    log("parse_event(..)=%s", pyev)
    if not pyev:
        return GDK_FILTER_CONTINUE  # @UndefinedVariable
    try:
        etype = pyev.type
        event_args = get_x_event_signals(etype)
        #log("signals(%s)=%s", pyev, event_args)
        if event_args is not None:
            signal, parent_signal = event_args
            _route_event(etype, pyev, signal, parent_signal)
        ename = get_x_event_type_name(etype) or etype
        log("x_event_filter event=%s/%s took %.1fms", event_args, ename, 1000.0*(monotonic()-start))
    except (KeyboardInterrupt, SystemExit):
        log("exiting on KeyboardInterrupt/SystemExit")
        Gtk.main_quit()
    except:
        log.warn("Unhandled exception in x_event_filter:", exc_info=True)
    return GDK_FILTER_CONTINUE  # @UndefinedVariable


_INIT_X11_FILTER_DONE = 0
def init_x11_filter() -> bool:
    log("init_x11_filter()")
    """ returns True if we did initialize it, False if it was already initialized """
    global _INIT_X11_FILTER_DONE
    cdef Display *display
    if _INIT_X11_FILTER_DONE==0:
        display = get_xdisplay()
        init_x11_events(display)
        gdk_window_add_filter(<GdkWindow*>0, x_event_filter, NULL)
    _INIT_X11_FILTER_DONE += 1
    return _INIT_X11_FILTER_DONE==1

def cleanup_x11_filter() -> bool:
    log("cleanup_x11_filter()")
    global _INIT_X11_FILTER_DONE
    _INIT_X11_FILTER_DONE -= 1
    if _INIT_X11_FILTER_DONE==0:
        gdk_window_remove_filter(<GdkWindow*>0, x_event_filter, NULL)
    return _INIT_X11_FILTER_DONE==0



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shlex
from subprocess import Popen, PIPE

from xpra.os_util import POSIX
from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS
from xpra.log import Logger

log = Logger("dbus")


def start_dbus(dbus_launch) -> tuple[int, dict]:
    if not dbus_launch or dbus_launch.lower() in FALSE_OPTIONS:
        log("start_dbus(%s) disabled", dbus_launch)
        return 0, {}
    if dbus_launch.lower() in TRUE_OPTIONS:
        log.warn(f"Warning: invalid dbus-launch command {dbus_launch!r}")
        return 0, {}
    bus_address = os.environ.get("DBUS_SESSION_BUS_ADDRESS")
    log("dbus_launch=%r, current DBUS_SESSION_BUS_ADDRESS=%s", dbus_launch, bus_address)
    if bus_address:
        log.warn("Warning: found an existing dbus instance:")
        log.warn(" DBUS_SESSION_BUS_ADDRESS=%s", bus_address)
    assert POSIX
    try:
        env = {k: v for k, v in os.environ.items() if k in (
            "PATH",
            "SSH_CLIENT", "SSH_CONNECTION",
            "XDG_CURRENT_DESKTOP", "XDG_SESSION_TYPE", "XDG_RUNTIME_DIR",
            "SHELL", "LANG", "USER", "LOGNAME", "HOME",
            "DISPLAY", "XAUTHORITY", "CKCON_X11_DISPLAY",
            "NO_AT_BRIDGE",
        )}
        cmd = shlex.split(dbus_launch)
        log("start_dbus(%s) env=%s", dbus_launch, env)
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, env=env, start_new_session=True, universal_newlines=True)
        out = proc.communicate()[0]
        assert proc.poll() == 0, "exit code is %s" % proc.poll()
        # parse and add to global env:
        dbus_env: dict[str, str] = {}
        log("out(%s)=%r", cmd, out)
        for line in out.splitlines():
            if line.startswith("export "):
                continue
            sep = "="
            if line.startswith("setenv "):
                line = line[len("setenv "):]
                sep = " "
            if line.startswith("set "):
                line = line[len("set "):]
            parts = line.split(sep, 1)
            if len(parts) != 2:
                continue
            k, v = parts
            if v.startswith("'") and v.endswith("';"):
                v = v[1:-2]
            elif v.endswith(";"):
                v = v[:-1]
            dbus_env[k] = v
        dbus_pid = int(dbus_env.get("DBUS_SESSION_BUS_PID", 0))
        log("dbus_pid=%i, dbus-env=%s", dbus_pid, dbus_env)
        return dbus_pid, dbus_env
    except Exception as e:
        log("start_dbus(%s)", dbus_launch, exc_info=True)
        log.error("dbus-launch failed to start using command '%s':\n" % dbus_launch)
        log.error(" %s\n" % e)
        return 0, {}



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import binascii
from typing import Any
from collections.abc import Callable, Iterable, Sequence

from xpra.common import noop
from xpra.platform import platform_import
from xpra.util.str_fn import bytestostr
from xpra.log import Logger

_init_done = False


def init() -> None:
    # warning: we currently call init() from multiple places to try
    # to ensure we run it as early as possible..
    global _init_done
    if not _init_done:
        _init_done = True
        do_init()


def do_init() -> None:
    """ some platforms override this """


_ready_done = False


def ready() -> None:
    global _ready_done
    if not _ready_done:
        _ready_done = True
        do_ready()


def do_ready() -> None:
    """ some platforms override this """


_default_icon = "xpra.png"


def set_default_icon(icon_filename: str) -> None:
    global _default_icon
    _default_icon = icon_filename


def get_default_icon():
    global _default_icon
    return _default_icon


def force_focus(duration=2000) -> None:
    # only implemented on macos
    assert isinstance(duration, int)


def use_stdin() -> bool:
    stdin = sys.stdin
    return bool(stdin) and stdin.isatty()


def get_clipboard_native_class() -> str:
    return ""


# defaults:
def get_native_tray_menu_helper_class() -> type | None:
    # classes that generate menus for xpra's system tray
    # let the toolkit classes use their own
    return None


def get_native_tray_classes(*_args) -> list[type]:
    # the classes we can use for our system tray:
    # let the toolkit classes use their own
    return []


def get_native_system_tray_classes(*_args) -> list[type]:
    # the classes we can use for application system tray forwarding:
    # let the toolkit classes use their own
    return []


def system_bell(*_args) -> bool:
    # let the toolkit classes use their own
    return False


def get_native_notifier_classes() -> list[type]:
    return []


def get_session_type() -> str:
    return ""


def get_xdpi() -> int:
    return -1


def get_ydpi() -> int:
    return -1


def get_monitors_info(xscale=1, yscale=1) -> dict[int, Any]:
    from xpra.gtk.info import get_monitors_info
    return get_monitors_info(xscale, yscale)


def get_icon_size() -> int:
    xdpi = get_xdpi()
    ydpi = get_ydpi()
    if xdpi > 0 and ydpi > 0:
        dpi = round((xdpi + ydpi) / 2)
    else:
        dpi = 96
    if dpi > 144:
        return 48
    if dpi > 120:
        return 32
    if dpi > 96:
        return 24
    return 16


def get_antialias_info() -> dict[str, Any]:
    return {}


def get_display_icc_info() -> dict[str, Any]:
    # per display info
    return {}


def get_icc_info() -> dict[str, Any]:
    return default_get_icc_info()


def default_get_icc_info() -> dict[str, Any]:
    ENV_ICC_DATA = os.environ.get("XPRA_ICC_DATA")
    if ENV_ICC_DATA:
        return {
            "source": "environment-override",
            "data": binascii.unhexlify(ENV_ICC_DATA),
        }
    return get_pillow_icc_info()


def get_pillow_icc_info() -> dict[str, Any]:
    screenlog = Logger("screen")
    info: dict[str, Any] = {}
    try:
        from PIL import ImageCms
        from PIL.ImageCms import get_display_profile
    except ImportError as e:
        screenlog.warn(f"Warning: unable to query color profile via Pillow: {e}")
        return info
    try:
        INTENT_STR: dict[Any, str] = {}
        for x in ("PERCEPTUAL", "RELATIVE_COLORIMETRIC", "SATURATION", "ABSOLUTE_COLORIMETRIC"):
            intent = getattr(ImageCms, "Intent", None)
            if intent:
                v = getattr(intent, x, None)
            else:
                v = getattr(ImageCms, "INTENT_%s" % x, None)
            if v:
                INTENT_STR[v] = x.lower().replace("_", "-")
        screenlog("get_icc_info() intents=%s", INTENT_STR)
        p = get_display_profile()  # NOSONAR @SuppressWarnings("python:S5727")
        screenlog("get_icc_info() display_profile=%s", p)
        if p:

            def getDefaultIntentStr(v) -> str:
                return INTENT_STR.get(v, "unknown")

            def getData(v) -> bytes:
                return v.tobytes()

            for (k, fn, conv) in (
                    ("name", "getProfileName", None),
                    ("info", "getProfileInfo", None),
                    ("copyright", "getProfileCopyright", None),
                    ("manufacturer", "getProfileManufacturer", None),
                    ("model", "getProfileModel", None),
                    ("description", "getProfileDescription", None),
                    ("default-intent", "getDefaultIntent", getDefaultIntentStr),
                    ("data", "getData", getData),
            ):
                m = getattr(ImageCms, fn, None)
                if m is None:
                    screenlog("%s lacks %s", ImageCms, fn)
                    continue
                try:
                    v = m(p)
                    if conv:
                        v = conv(v)
                    info[k] = bytestostr(v).rstrip("\n\r")
                except Exception as e:
                    screenlog("get_icc_info()", exc_info=True)
                    screenlog("ICC profile error on %s using %s: %s", k, fn, e)
    except Exception as e:
        screenlog("get_icc_info()", exc_info=True)
        screenlog.warn("Warning: cannot query ICC profiles:")
        screenlog.warn(" %s", e)
    return info


# global workarea for all screens
def get_workarea() -> tuple[int, int, int, int] | None:
    return None


# per monitor workareas (assuming a single screen)
def get_workareas() -> Sequence[tuple[int, int, int, int]]:
    return ()


def get_number_of_desktops() -> int:
    return 1


def get_desktop_names() -> Sequence[str]:
    return ()


def get_vrefresh() -> int:
    return -1


def get_mouse_config() -> dict:
    return {}


def get_double_click_time() -> int:
    return -1


def get_double_click_distance() -> tuple[int, int]:
    return -1, -1


def get_fixed_cursor_size() -> tuple[int, int]:
    return -1, -1


def get_cursor_size() -> int:
    return -1


def get_window_min_size() -> tuple[int, int]:
    return 0, 0


def get_window_max_size() -> tuple[int, int]:
    return 2 ** 15 - 1, 2 ** 15 - 1


def get_window_frame_size(_x, _y, _w, _h):
    return None


def get_window_frame_sizes() -> dict[str, Any]:
    return {}


def add_window_hooks(_window) -> None:
    """
    To add platform specific code to each window,
    called when the window is created.
    """


def remove_window_hooks(_window) -> None:
    """
    Remove the hooks,
    called when the window is destroyed
    """


def show_desktop(_show) -> None:
    """ If possible, show the desktop """


def set_fullscreen_monitors(_window, _fsm, _source_indication: int = 0) -> None:
    """ Only overridden by posix """


def set_shaded(_window, _shaded: bool) -> None:
    """
    GTK never exposed the 'shaded' window attribute,
    posix clients will hook it up here.
    """


def pointer_grab(_window):
    """
    Pointer grabs require platform specific code
    """
    return False


def pointer_ungrab(_window):
    """
    Pointer grabs require platform specific code
    """
    return False


def gl_check() -> str:
    return ""  # no problem


def get_wm_name() -> str:
    return ""


def can_access_display() -> bool:
    return True


def set_window_progress(window, pct: int) -> None:
    """ some platforms can indicate progress for a specific window """


take_screenshot: Callable = noop
ClientExtras = None


def get_info_base() -> dict[str, Any]:
    def fname(v):
        try:
            return v.__name__
        except AttributeError:
            return str(v)

    def fnames(flist: Iterable) -> Sequence:
        return [fname(x) for x in flist]

    return {
        "native-clipboard": get_clipboard_native_class(),
        "native_tray_menu_helper": fname(get_native_tray_menu_helper_class()),
        "native_trays": fnames(get_native_tray_classes()),
        "native_system_trays": fnames(get_native_system_tray_classes()),
        "system_bell": fname(system_bell),
        "native_notifiers": fnames(get_native_notifier_classes()),
        "wm_name": get_wm_name() or "",
        "workarea": get_workarea() or "",
        "workareas": get_workareas(),
        "monitors": get_monitors_info(),
        "desktops": get_number_of_desktops(),
        "desktop_names": get_desktop_names(),
        "session-type": get_session_type(),
        "vertical-refresh": get_vrefresh(),
        "fixed_cursor_size": get_fixed_cursor_size(),
        "cursor_size": get_cursor_size(),
        "icon_size": get_icon_size(),
        "mouse": get_mouse_config(),
        "double_click": {
            "time": get_double_click_time(),
            "distance": get_double_click_distance(),
        },
        "dpi": {
            "x": get_xdpi(),
            "y": get_ydpi(),
        },
        "antialias": get_antialias_info(),
        "icc": get_icc_info(),
        "display-icc": get_display_icc_info(),
        "window_frame": get_window_frame_sizes(),
        "can_access_display": can_access_display(),
    }


get_info = get_info_base

platform_import(globals(), "gui", False,
                "do_ready",
                "do_init",
                "force_focus",
                "gl_check",
                "use_stdin",
                "get_wm_name",
                "show_desktop", "set_fullscreen_monitors", "set_shaded",
                "pointer_grab", "pointer_ungrab",
                "ClientExtras",
                "take_screenshot",
                "get_clipboard_native_class",
                "get_native_tray_menu_helper_class",
                "get_native_tray_classes",
                "get_native_system_tray_classes",
                "get_native_notifier_classes",
                "get_session_type",
                "get_vrefresh", "get_workarea", "get_workareas",
                "get_number_of_desktops", "get_desktop_names",
                "get_antialias_info", "get_icc_info", "get_display_icc_info", "get_xdpi", "get_ydpi",
                "get_monitors_info",
                "get_icon_size",
                "get_window_min_size", "get_window_max_size",
                "get_mouse_config",
                "get_double_click_time", "get_double_click_distance",
                "get_fixed_cursor_size", "get_cursor_size", "get_window_frame_sizes",
                "add_window_hooks", "remove_window_hooks",
                "system_bell",
                "can_access_display",
                "set_window_progress",
                "get_info",
                )


def main() -> int:
    from xpra.platform import program_context
    from xpra.util.str_fn import print_nested_dict
    from xpra.os_util import OSX, POSIX
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("GUI-Properties"):
        enable_color()
        consume_verbose_argv(sys.argv, "all")
        init()

        # naughty, but how else can I hook this up?
        if POSIX and not OSX:
            from xpra.x11.bindings.posix_display_source import init_posix_display_source
            init_posix_display_source()
        i = get_info()
        print_nested_dict(i, hex_keys=("data", "icc-data", "icc-profile"))
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.platform.win32.auth import check
from xpra.auth.sys_auth_base import SysAuthenticator, log


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # fugly: keep hold of the password so the win32 proxy can use it
        self.password = ""

    def verify_username(self, remote_username: str) -> None:
        if remote_username.lower() != self.username.lower():
            log(f"verifying username={self.username!r} vs remote={remote_username!r}")
            raise ValueError(f"invalid username {remote_username!r}")

    def get_uid(self) -> int:
        return 0

    def get_gid(self) -> int:
        return 0

    def get_password(self) -> str:
        return self.password

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        self.req_xor(digests)
        return super().do_get_challenge(["xor"])

    def check_password(self, password: str) -> bool:
        domain = ""  # os.environ.get('COMPUTERNAME')
        if check(domain, self.username, password):
            self.password = password
            return True
        return False

    def __repr__(self):
        return "win32"


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("Auth-Test", "Auth-Test"):
        enable_color()
        consume_verbose_argv(argv, "auth")
        if len(argv) != 3:
            log.warn("invalid number of arguments")
            log.warn("usage: %s [--verbose] username password", argv[0])
            return 1
        username = argv[1]
        password = argv[2]
        if check("", username, password):
            log.info("authentication succeeded")
            return 0
        log.error("authentication failed")
        return 1


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.common import noop
from xpra.util.objects import AdHocStruct
from xpra.client.gui.keyboard_helper import KeyboardHelper
from unit.process_test_util import DisplayContext


class KeyboardHelperTest(unittest.TestCase):

    def test_modifier(self):
        kh = KeyboardHelper(noop)

        def checkmask(mask, *modifiers):
            #print("checkmask(%s, %s)", mask, modifiers)
            mods = kh.mask_to_names(mask)
            assert set(mods) == set(modifiers), "expected %s got %s" % (modifiers, mods)

        from gi.repository import Gdk  # @UnresolvedImport
        checkmask(Gdk.ModifierType.SHIFT_MASK, "shift")
        checkmask(Gdk.ModifierType.LOCK_MASK, "lock")
        if getattr(kh.keyboard, "swap_keys", False):
            checkmask(Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.META_MASK, "shift", "control")
            #turn swap off and run again:
            kh.keyboard.swap_keys = False
        checkmask(Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK, "shift", "control")
        kh.cleanup()

    def test_keymap_properties(self):
        kh = KeyboardHelper(noop)
        kh.query_xkbmap()
        p = kh.get_keymap_properties()
        assert p, "no keymap properties returned"
        assert len(p) >= 8, "not enough keymap properties (%i): %s" % (len(p), p)
        kh.cleanup()

    def test_parse_shortcuts(self):
        shortcuts = [
            'Control+Menu:toggle_keyboard_grab',
            'Shift+Menu:toggle_pointer_grab',
            'Shift+F11:toggle_fullscreen',
            '#+F1:show_menu',
            'Control+F1:show_window_menu',
            '#+F2:show_start_new_command',
            '#+F3:show_bug_report',
            '#+F4:quit',
            '#+F5:increase_quality',
            '#+F6:decrease_quality',
            '#+F7:increase_speed',
            '#+F8:decrease_speed',
            '#+F10:magic_key',
            '#+F11:show_session_info',
            '#+F12:toggle_debug',
            '#+plus:scaleup',
            '#+minus:scaledown',
            '#+underscore:scaledown',
            '#+KP_Add:scaleup',
            '#+KP_Subtract:scaledown',
            '#+KP_Multiply:scalereset',
            '#+bar:scalereset',
            '#+question:scalingoff',
        ]
        kh = KeyboardHelper(noop, key_shortcuts=shortcuts)
        parsed = kh.parse_shortcuts()
        assert kh.shortcut_modifiers, "no shortcut modifiers: %s" % (kh.shortcut_modifiers,)
        assert len(parsed) > 10, "not enough shortcuts parsed: %s" % (parsed,)
        window = AdHocStruct()
        window.quit = noop
        modifier_names = kh.get_modifier_names()
        modifiers_used = [modifier_names.get(x, x) for x in kh.shortcut_modifiers]
        assert kh.key_handled_as_shortcut(window, "F4", modifiers_used, True)
        assert not kh.key_handled_as_shortcut(window, "F1", [], True)
        kh.cleanup()


def main():
    with DisplayContext():
        unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
import signal
import unittest
from subprocess import Popen, PIPE

from xpra.util.env import envbool
from xpra.os_util import POSIX, OSX
from xpra.util.io import pollwait
from unit.process_test_util import ProcessTestUtil, log


USE_DISPLAY = envbool("XPRA_TEST_USE_DISPLAY", False)


class SplashTest(ProcessTestUtil):

    def setUp(self):
        ProcessTestUtil.setUp(self)
        self.splash = None
        self.display = None
        self.xvfb = None
        if not POSIX or OSX:
            return
        if USE_DISPLAY:
            self.display = os.environ.get("DISPLAY", "")
            return
        client_display = self.find_free_display()
        self.xvfb = self.start_Xvfb(client_display)
        self.display = client_display
        try:
            from xpra.x11.bindings.wait_for_x_server import wait_for_x_server
        except ImportError:
            time.sleep(5)
        else:
            wait_for_x_server(client_display, 10)

    def tearDown(self):
        self.stop_splash()
        self.stop_vfb()

    def stop_vfb(self):
        xvfb = self.xvfb
        if xvfb:
            self.xvfb = None
            xvfb.terminate()

    def stop_splash(self):
        s = self.splash
        if not s:
            return
        try:
            self.do_feed_splash(s, [
                "100:100",
            ], 5)
            if pollwait(s, 5) is not None:
                return
            s.terminate()
            if pollwait(s, 5) is not None:
                return
            try:
                s.kill()
            except Exception:
                pass
        finally:
            try:
                s.stdin.close()
            except Exception:
                pass
            self.splash = None

    def _run_splash(self):
        env = self.get_run_env()
        env["XPRA_SPLASH_FOCUS_EXIT"] = "0"
        if self.display:
            env["DISPLAY"] = self.display
        cmd = self.get_xpra_cmd()
        cmd += ["splash"]
        log("_run_splash() env=%s, cmd=%s", env, cmd)
        self.splash = Popen(args=cmd, stdin=PIPE, env=env, start_new_session=True)
        return self.splash

    def _feed_splash(self, lines=None, delay=1):
        proc = self._run_splash()
        return self.do_feed_splash(proc, lines, delay)

    def do_feed_splash(self, proc, lines=None, delay=1):
        while lines and proc.poll() is None:
            line = lines.pop(0)
            log("sending '%s'", line)
            proc.stdin.write(line.encode()+b"\n\r")
            proc.stdin.flush()
            if pollwait(proc, delay) is not None:
                break
        return proc

    def test_invalid(self):
        self._feed_splash([
            "notanumber:ignoreit",
            "50:50",
        ])
        r = pollwait(self.splash, 5)
        assert r is None, "splash screen should not have terminated"
        #try killing it with a signal:
        self.splash.send_signal(signal.SIGTERM)
        r = pollwait(self.splash, 20)
        assert r is not None, "expected splash to exit"

    def test_full(self):
        self._feed_splash([
            "10:10",
            "100:100",
        ])
        r = pollwait(self.splash, 20)
        assert r is not None, "splash screen should have terminated"
        assert r==0, "exit code should be zero, but got %s" % r
        self.stop_splash()

    def test_partial(self):
        self._feed_splash([
            "10:10",
            "20:20",
            "90:90",
        ])
        assert self.splash.poll() is None, "splash screen should still be running"
        self.stop_splash()


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.os_util import gi_import
from xpra.util.screen import log_screen_sizes
from xpra.util.str_fn import csv
from xpra.util.env import envbool
from xpra.net.common import PacketType
from xpra.server import features
from xpra.gtk.util import get_root_size
from xpra.gtk.info import get_screen_sizes
from xpra.gtk.gobject import one_arg_signal
from xpra.x11.gtk.bindings import add_catchall_receiver, remove_catchall_receiver, add_event_receiver
from xpra.x11.xroot_props import XRootPropWatcher
from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.bindings.keyboard import X11KeyboardBindings
from xpra.x11.server.base import X11ServerBase
from xpra.gtk.error import xsync, xlog
from xpra.log import Logger

GObject = gi_import("GObject")
Gio = gi_import("Gio")

X11Window = X11WindowBindings()
X11Keyboard = X11KeyboardBindings()

log = Logger("server")
windowlog = Logger("server", "window")
mouselog = Logger("server", "mouse")
geomlog = Logger("server", "window", "geometry")
metadatalog = Logger("x11", "metadata")
screenlog = Logger("screen")
iconlog = Logger("icon")

MODIFY_GSETTINGS: bool = envbool("XPRA_MODIFY_GSETTINGS", True)
MULTI_MONITORS: bool = envbool("XPRA_DESKTOP_MULTI_MONITORS", True)


def get_desktop_server_base_classes() -> tuple[type, ...]:
    classes: list[type] = [GObject.GObject]
    if features.rfb:
        from xpra.server.rfb.server import RFBServer
        classes.append(RFBServer)
    classes.append(X11ServerBase)
    return tuple(classes)


DESKTOPSERVER_BASES = get_desktop_server_base_classes()
DesktopServerBaseClass = type('DesktopServerBaseClass', DESKTOPSERVER_BASES, {})
log("DesktopServerBaseClass%s", DESKTOPSERVER_BASES)


def do_modify_gsettings(defs: dict[str, Any], value=False) -> dict[str, Any]:
    modified = {}
    try:
        schemas = Gio.SettingsSchemaSource.get_default().list_schemas(True)
    except AttributeError:
        schemas = Gio.Settings.list_schemas()
    for schema, attributes in defs.items():
        if schema not in schemas:
            continue
        try:
            s = Gio.Settings.new(schema_id=schema)
            restore = []
            for attribute in attributes:
                v = s.get_boolean(attribute)
                if v:
                    s.set_boolean(attribute, value)
                    restore.append(attribute)
            if restore:
                modified[schema] = restore
        except Exception as e:
            log("error accessing schema '%s' and attributes %s", schema, attributes, exc_info=True)
            log.error("Error accessing schema '%s' and attributes %s:", schema, csv(attributes))
            log.estr(e)
    return modified


class DesktopServerBase(DesktopServerBaseClass):
    """
        A server base class for RFB / VNC-like virtual desktop or virtual monitors,
        used with the "start-desktop" subcommand.
    """
    __common_gsignals__: dict[str, tuple] = {
        "x11-xkb-event": one_arg_signal,
        "x11-cursor-event": one_arg_signal,
        "x11-motion-event": one_arg_signal,
        "x11-configure-event": one_arg_signal,
    }

    def __init__(self):
        X11ServerBase.__init__(self)  # pylint: disable=non-parent-init-called
        for c in DESKTOPSERVER_BASES:
            if c != X11ServerBase:
                c.__init__(self)  # pylint: disable=non-parent-init-called
        self.gsettings_modified: dict[str, Any] = {}
        self.root_prop_watcher = None

    def init(self, opts) -> None:
        for c in DESKTOPSERVER_BASES:
            if c != GObject.GObject:
                c.init(self, opts)

    def x11_init(self) -> None:
        X11ServerBase.x11_init(self)
        add_event_receiver(X11Window.get_root_xid(), self)
        add_catchall_receiver("x11-motion-event", self)
        add_catchall_receiver("x11-xkb-event", self)
        with xlog:
            X11Keyboard.selectBellNotification(True)
        if MODIFY_GSETTINGS:
            self.modify_gsettings()
        self.root_prop_watcher = XRootPropWatcher(["WINDOW_MANAGER", "_NET_SUPPORTING_WM_CHECK"])
        self.root_prop_watcher.connect("root-prop-changed", self.root_prop_changed)

    def root_prop_changed(self, watcher, prop: str) -> None:
        iconlog("root_prop_changed(%s, %s)", watcher, prop)
        for window in self._id_to_window.values():
            window.update_wm_name()
            window.update_icon()

    def modify_gsettings(self) -> None:
        # try to suspend animations:
        self.gsettings_modified = do_modify_gsettings({
            "org.mate.interface": ("gtk-enable-animations", "enable-animations"),
            "org.gnome.desktop.interface": ("enable-animations",),
            "com.deepin.wrap.gnome.desktop.interface": ("enable-animations",),
        })

    def do_cleanup(self) -> None:
        remove_catchall_receiver("x11-motion-event", self)
        X11ServerBase.do_cleanup(self)
        if MODIFY_GSETTINGS:
            self.restore_gsettings()
        rpw = self.root_prop_watcher
        if rpw:
            self.root_prop_watcher = None
            rpw.cleanup()

    def restore_gsettings(self) -> None:
        do_modify_gsettings(self.gsettings_modified, True)

    def notify_dpi_warning(self, body) -> None:
        """ ignore DPI warnings in desktop mode """

    def print_screen_info(self) -> None:
        super().print_screen_info()
        root_w, root_h = get_root_size()
        log.info(" initial resolution: %ix%i", root_w, root_h)
        sss = get_screen_sizes()
        log_screen_sizes(root_w, root_h, sss)

    def parse_screen_info(self, ss) -> None:
        return self.do_parse_screen_info(ss, ss.desktop_mode_size)

    def notify_screen_changed(self, screen) -> None:
        """
        Screen changes are normally managed by requests or user actions,
        we do not need to send any messages to the client here,
        the monitor window model(s) will take care of it.
        """

    def set_desktop_geometry_attributes(self, w: int, h: int):
        # geometry is not synced with the client's for desktop servers
        pass

    def get_server_mode(self) -> str:
        return "X11 desktop"

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        if "features" in source.wants:
            capabilities.setdefault("pointer", {})["grabs"] = True
            capabilities["desktop"] = True
            capabilities.setdefault("window", {}).update({
                "decorations": True,
                "states": ["iconified", "focused"],
            })
            capabilities["screen_sizes"] = get_screen_sizes()
        return capabilities

    def load_existing_windows(self):
        raise NotImplementedError

    def send_initial_windows(self, ss, sharing: bool = False) -> None:
        windowlog("send_initial_windows(%s, %s) will send: %s", ss, sharing, self._id_to_window)
        for model in self._id_to_window.values():
            self.send_new_desktop_model(model, ss, sharing)

    def send_new_desktop_model(self, model, ss, _sharing: bool = False) -> None:
        x, y, w, h = model.get_geometry()
        wid = self._window_to_id[model]
        wprops = self.client_properties.get(wid, {}).get(ss.uuid, {})
        ss.new_window("new-window", wid, model, x, y, w, h, wprops)
        wid = self._window_to_id[model]
        ss.damage(wid, model, 0, 0, w, h)

    def _lost_window(self, window, wm_exiting=False) -> None:
        """ could be used to slow down the refresh rate? """

    def _contents_changed(self, window, event) -> None:
        log("contents changed on %s: %s", window, event)
        self.refresh_window_area(window, event.x, event.y, event.width, event.height)

    def _set_window_state(self, proto, wid: int, window, new_window_state) -> list[str]:
        if not new_window_state:
            return []
        metadatalog("set_window_state%s", (proto, wid, window, new_window_state))
        changes = []
        # boolean: but not a wm_state and renamed in the model... (iconic vs iconified!)
        iconified = new_window_state.get("iconified")
        if iconified is not None and window._updateprop("iconic", iconified):
            changes.append("iconified")
        focused = new_window_state.get("focused")
        if focused is not None and window._updateprop("focused", focused):
            changes.append("focused")
        return changes

    def get_window_position(self, _window) -> tuple[int, int]:
        # we export the whole desktop as a window:
        return 0, 0

    def _process_map_window(self, proto, packet: PacketType) -> None:
        wid, x, y, w, h = (int(x) for x in packet[1:6])
        window = self._id_to_window.get(wid)
        if not window:
            windowlog("cannot map window %s: already removed!", wid)
            return
        geomlog("client mapped window %s - %s, at: %s", wid, window, (x, y, w, h))
        self._window_mapped_at(proto, wid, window, (x, y, w, h))
        if len(packet) >= 8:
            self._set_window_state(proto, wid, window, packet[7])
        if len(packet) >= 7:
            self._set_client_properties(proto, wid, window, packet[6])
        self.refresh_window_area(window, 0, 0, w, h)

    def _process_unmap_window(self, proto, packet: PacketType) -> None:
        wid = int(packet[1])
        window = self._id_to_window.get(wid)
        if not window:
            log("cannot map window %s: already removed!", wid)
            return
        if len(packet) >= 4:
            # optional window_state added in 0.15 to update flags
            # during iconification events:
            self._set_window_state(proto, wid, window, packet[3])
        assert not window.is_OR()
        self._window_mapped_at(proto, wid, window)
        # TODO: handle inconification?
        # iconified = len(packet)>=3 and bool(packet[2])

    def _process_configure_window(self, proto, packet: PacketType) -> None:
        wid, x, y, w, h = (int(x) for x in packet[1:6])
        if len(packet) >= 13 and features.input_devices and not self.readonly:
            pwid = int(packet[10])
            pointer = packet[11]
            modifiers = packet[12]
            device_id = -1
            if self.process_mouse_common(proto, device_id, pwid, pointer):
                self._update_modifiers(proto, wid, modifiers)
        # some "configure-window" packets are only meant for metadata updates:
        skip_geometry = len(packet) >= 10 and packet[9]
        window = self._id_to_window.get(wid)
        if not window:
            geomlog("cannot map window %s: already removed!", wid)
            return
        damage = False
        if len(packet) >= 9:
            damage = bool(self._set_window_state(proto, wid, window, packet[8]))
        if not skip_geometry and not self.readonly:
            owx, owy, oww, owh = window.get_geometry()
            geomlog("_process_configure_window(%s) old window geometry: %s", packet[1:], (owx, owy, oww, owh))
            if oww != w or owh != h:
                window.resize(w, h)
        if len(packet) >= 7:
            cprops = packet[6]
            if cprops:
                metadatalog("window client properties updates: %s", cprops)
                self._set_client_properties(proto, wid, window, cprops)
        self._window_mapped_at(proto, wid, window, (x, y, w, h))
        if damage:
            self.refresh_window_area(window, 0, 0, w, h)

    def _adjust_pointer(self, proto, device_id: int, wid: int, pointer):
        mouselog("_adjust_pointer%s", (proto, device_id, wid, pointer))
        window = self._id_to_window.get(wid)
        if not window:
            mouselog("adjust pointer: no window, suspending cursor")
            self.suspend_cursor(proto)
            return None
        pointer = super()._adjust_pointer(proto, device_id, wid, pointer)
        # maybe the pointer is off-screen:
        ww, wh = window.get_dimensions()
        x, y = pointer[:2]
        if x < 0 or x >= ww or y < 0 or y >= wh:
            mouselog("adjust pointer: pointer outside desktop, suspending cursor")
            self.suspend_cursor(proto)
            return None
        self.restore_cursor(proto)
        return pointer

    def _move_pointer(self, device_id: int, wid: int, pos, props=None):
        if wid >= 0:
            window = self._id_to_window.get(wid)
            if not window:
                mouselog("_move_pointer(%s, %s) invalid window id", wid, pos)
                return
        with xsync:
            X11ServerBase._move_pointer(self, device_id, wid, pos, props)

    def _process_close_window(self, proto, packet: PacketType) -> None:
        # disconnect?
        pass

    def _process_desktop_size(self, proto, packet: PacketType) -> None:
        """
        Usually, desktop servers don't need to do anything when the client's geometry changes.
        """

    def calculate_workarea(self, w: int, h: int):
        """
        The workarea is managed server side by the window manager,
        so we don't need to apply any changes here
        """

    def make_dbus_server(self) -> None:
        from xpra.x11.dbus.x11_dbus_server import X11_DBUS_Server
        self.dbus_server = X11_DBUS_Server(self, os.environ.get("DISPLAY", "").lstrip(":"))

    def show_all_windows(self) -> None:
        log.warn("Warning: show_all_windows not implemented for desktop server")

    def do_make_screenshot_packet(self) -> tuple:
        log("grabbing screenshot")
        regions = []
        offset_x, offset_y = 0, 0
        for wid in reversed(sorted(self._id_to_window.keys())):
            window = self._id_to_window.get(wid)
            log("screenshot: window(%s)=%s", wid, window)
            if window is None:
                continue
            if not window.is_managed():
                log("screenshot: window %s is not/no longer managed", wid)
                continue
            x, y, w, h = window.get_geometry()
            log("screenshot: geometry(%s)=%s", window, (x, y, w, h))
            try:
                with xsync:
                    img = window.get_image(0, 0, w, h)
            except Exception:
                log.warn("screenshot: window %s could not be captured", wid)
                continue
            if img is None:
                log.warn("screenshot: no pixels for window %s", wid)
                continue
            log("screenshot: image=%s, size=%s", img, img.get_size())
            if img.get_pixel_format() not in ("RGB", "RGBA", "XRGB", "BGRX", "ARGB", "BGRA"):
                log.warn("window pixels for window %s using an unexpected rgb format: %s", wid, img.get_pixel_format())
                continue
            regions.append((wid, offset_x + x, offset_y + y, img))
            # tile them horizontally:
            offset_x += w
            offset_y += 0
        return self.make_screenshot_packet_from_regions(regions)



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shlex
from typing import Any
from collections.abc import Callable

from xpra.util.objects import typedict
from xpra.os_util import WIN32


class StubServerMixin:
    """
    Base class for server mixins.
    Defines the default interface methods that each mixin may override.
    """

    def init(self, _opts):
        """
        Initialize this instance with the options given.
        Options are usually obtained by parsing the command line,
        or using a default configuration object.
        """

    def init_state(self) -> None:
        """
        Initialize state attributes.
        """

    def add_init_thread_callback(self, callback: Callable) -> None:
        """
        Adds a callback that will be executed
        after the init thread has completed.
        """

    def last_client_exited(self) -> None:
        """
        Called when the last client has exited,
        so we can reset things to their original state.
        """

    def cleanup(self) -> None:
        """
        Free up any resources.
        """

    def late_cleanup(self, stop=True) -> None:
        """
        Free up any resources, after main cleanup.
        `stop` is set to True when we are meant to stop all subprocesses we are responsible for.
        (ie: `Xvfb` and such)
        """

    def setup(self) -> None:
        """
        After initialization, prepare to run.
        """

    def threaded_setup(self) -> None:
        """
        Prepare to run, this method runs in parallel to start faster.
        """

    def init_sockets(self, _sockets) -> None:
        """
        Prepare to handle connections from the given sockets.
        """

    def get_caps(self, _source) -> dict[str, Any]:
        """
        Capabilities provided by this mixin.
        """
        return {}

    def get_server_features(self, _source) -> dict[str, Any]:
        """
        Features provided by this mixin.
        (the difference with capabilities is that those will only
        be returned if the client requests 'features')
        """
        return {}

    def set_session_driver(self, _source) -> None:
        """
        When the user in control of the session changes,
        this method will be called.
        """

    def get_info(self, _proto) -> dict[str, Any]:
        """
        Runtime information on this mixin, includes state and settings.
        Somewhat overlaps with the capabilities and features,
        but the data is returned in a structured format. (ie: nested dictionaries)
        """
        return {}

    def get_ui_info(self, proto, client_uuids=None, *args) -> dict[str, Any]:
        """
        Runtime information on this mixin,
        unlike get_info() this method will be called
        from the UI thread.
        """
        return {}

    def init_packet_handlers(self) -> None:
        """
        Register the packet types that this mixin can handle.
        """

    def parse_hello(self, ss, caps: typedict, send_ui: bool) -> None:
        """
        Parse capabilities from a new connection.
        """

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        """
        A new client is being handled, take any action needed.
        """

    def send_initial_data(self, ss, caps, send_ui, share_count: int) -> None:
        """
        A new connection has been accepted, send initial data.
        """

    def cleanup_protocol(self, protocol) -> None:
        """
        Cleanup method for a specific connection.
        (to clean up / free up resources associated with a specific client or connection)
        """

    def get_child_env(self) -> dict[str, str]:
        return os.environ.copy()

    def get_full_child_command(self, cmd, _use_wrapper: bool = True) -> list[str]:
        # make sure we have it as a list:
        if isinstance(cmd, (list, tuple)):
            return list(cmd)
        if WIN32:  # pragma: no cover
            return [cmd]
        return shlex.split(str(cmd))

    def get_server_source(self, _proto):
        """ returns the client connection source object for the given protocol """
        return None



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>

import re
import sys
from collections import deque
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.util.str_fn import csv
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.platform.features import CLIPBOARDS

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")


class ClipboardInstance:
    def __init__(self, selection, _log):
        atom = Gdk.Atom.intern(selection, False)
        self.clipboard = Gtk.Clipboard.get(atom)
        self.selection = selection
        self._log = _log
        self.owned_label = label()
        self.get_targets = Gtk.ComboBoxText()
        self.get_targets.set_sensitive(False)
        self.get_targets.connect("changed", self.get_target_changed)
        self.set_targets = Gtk.ComboBoxText()
        self.set_targets.append_text("STRING")
        self.set_targets.append_text("UTF8_STRING")
        self.set_targets.set_active(0)
        self.set_targets.connect("changed", self.set_target_changed)
        self.value_label = label()
        self.value_entry = Gtk.Entry()
        self.value_entry.set_max_length(100)
        self.value_entry.set_width_chars(32)

        def b(text: str, callback: Callable):
            btn = Gtk.Button(label=text)
            btn.connect("clicked", callback)
            return btn

        self.clear_label_btn = b("X", self.clear_label)
        self.clear_entry_btn = b("X", self.clear_entry)
        self.get_get_targets_btn = b("Get Targets", self.do_get_targets)
        self.get_target_btn = b("Get Target", self.do_get_target)
        self.get_target_btn.set_sensitive(False)
        self.set_target_btn = b("Set Target", self.do_set_target)
        self.get_string_btn = b("Get String", self.do_get_string)
        self.set_string_btn = b("Set String", self.do_set_string)
        self.clipboard.connect("owner-change", self.owner_changed)
        self.log("ready")

    def __repr__(self):
        return "ClipboardInstance(%s)" % self.selection

    def log(self, msg):
        self._log(self.selection, msg)

    def clear_entry(self, *_args):
        self.value_entry.set_text("")

    def clear_label(self, *_args):
        self.value_label.set_text("")

    def get_targets_callback(self, _c, targets, *_args):
        self.log("got targets: %s" % csv(str(x) for x in targets))
        if hasattr(targets, "name"):
            self.log("target is atom: %s" % targets.name())
            targets = []
        filtered = [x for x in (targets or []) if x not in ("MULTIPLE", "TARGETS")]
        ct = self.get_targets.get_active_text()
        if not ct:
            # choose a good default target:
            for x in ("STRING", "UTF8_STRING"):
                if x in filtered:
                    ct = x
                    break
        self.get_targets.get_model().clear()
        self.get_targets.set_sensitive(True)
        i = 0
        for t in filtered:
            self.get_targets.append_text(str(t))
            if t == ct:
                self.get_targets.set_active(i)
            i += 1
        self.get_targets.show_all()

    def do_get_targets(self, *_args):
        self.clipboard.request_targets(self.get_targets_callback, None)

    def get_target_changed(self, _cb):
        target = self.get_targets.get_active_text()
        self.get_target_btn.set_sensitive(bool(target))

    def set_target_changed(self, cb):
        self.log("set_target_changed(%s) target=%s" % (cb, self.set_targets.get_active_text()))

    def ellipsis(self, val):
        if len(val) > 24:
            return val[:24] + ".."
        return val

    def selection_value_callback(self, _cb, selection_data, *_args):
        data = b""
        try:
            data = selection_data.get_data()
            if data is None:
                s = ""
            else:
                s = "type=%s, format=%s, data=%s" % (
                    selection_data.get_data_type(),
                    selection_data.get_format(),
                    self.ellipsis(re.escape(data)))
        except TypeError:
            try:
                s = self.ellipsis("\\".join([str(x) for x in bytearray(data)]))
            except Exception as e:
                s = f"!ERROR: {e}! binary data?"
        self.log("Got selection data: '%s'" % s)
        self.value_label.set_text(s)

    def do_get_target(self, *_args):
        self.clear_label()
        target = self.get_targets.get_active_text()
        self.log("Requesting %s" % target)
        atom = Gdk.Atom.intern(target, False)
        self.clipboard.request_contents(atom, self.selection_value_callback, None)

    def selection_clear_cb(self, _clipboard, _data):
        self.log("Selection has been cleared")

    def selection_get_callback(self, _clipboard, selectiondata, _info, *_args):
        # log("selection_get_callback(%s, %s, %s, %s) targets=%s",
        #    clipboard, selectiondata, info, args, selectiondata.get_targets())
        value = self.value_entry.get_text()
        self.log("Answering selection request with value: '%s'" % self.ellipsis(value))
        selectiondata.set("STRING", 8, value)

    def do_set_target(self, *_args):
        target = self.set_targets.get_active_text()
        self.log("Target set to %s" % target)
        self.clipboard.set_with_data([(target, 0, 0)], self.selection_get_callback, self.selection_clear_cb)

    def string_value_callback(self, _cb, value, *_args):
        if value is None:
            value = ""
        assert isinstance(value, str), "value is not a string!"
        self.log("Got string selection data: '%s'" % value)
        self.value_label.set_text(self.ellipsis(value))

    def do_get_string(self, *_args):
        # self.log("do_get_string%s on %s.%s" % (args, self, self.clipboard))
        self.clipboard.request_text(self.string_value_callback, None)

    def do_set_string(self, *_args):
        self.clipboard.set_text(self.ellipsis(self.value_entry.get_text()))

    def owner_changed(self, _cb, event):
        owner = self.clipboard.get_owner()
        weownit = (owner is not None)
        if weownit:
            owner_info = "(us)"
        else:
            owner_info = str(event.owner or 0)
        self.log("Owner changed, reason: %s, new owner=%s" % (event.reason, owner_info))


class ClipboardStateInfoWindow:

    def __init__(self):
        self.window = Gtk.Window()
        self.window.connect("destroy", self.destroy)
        self.window.set_default_size(640, 300)
        self.window.set_border_width(20)
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.set_title("Clipboard Test Tool")

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(15)

        self.log: deque[str] = deque(maxlen=25)
        for x in range(25):
            self.log.append("")
        self.events = label("", font="monospace 9")

        # how many clipboards to show:
        self.clipboards = CLIPBOARDS

        grid = Gtk.Grid()
        for i, text in enumerate(("Selection", "Value", "Clear", "Targets", "Actions")):
            grid.attach(label(text), i, 1, 1, 1)

        for row, selection in enumerate(self.clipboards):
            grid.attach(label(selection), 0, 2 + row * 2, 1, 2)
            cs = ClipboardInstance(selection, self.add_event)
            get_actions = Gtk.HBox()
            for x in (cs.get_get_targets_btn, cs.get_target_btn, cs.get_string_btn):
                get_actions.pack_start(x)
            for i, widget in enumerate((cs.value_label, cs.clear_label_btn, cs.get_targets, get_actions)):
                grid.attach(widget, 1 + i, 2 + row * 2, 1, 1)
            set_actions = Gtk.HBox()
            for x in (cs.set_target_btn, cs.set_string_btn):
                set_actions.pack_start(x)
            widgets = (cs.value_entry, cs.clear_entry_btn, cs.set_targets, set_actions)
            for i, widget in enumerate(widgets):
                grid.attach(widget, 1 + i, 3 + row * 2, 1, 1)
        vbox.pack_start(grid)
        vbox.add(self.events)

        self.window.add(vbox)
        self.window.show_all()
        icon = get_icon_pixbuf("clipboard.png")
        if icon:
            self.window.set_icon(icon)
        try:
            self.add_event("ALL", "window=%s, xid=%#x" % (self.window, self.window.get_window().get_xid()))
        except Exception:
            self.add_event("ALL", "window=%s" % self.window)

    def add_event(self, selection, message):
        msg = message
        if self.clipboards:
            msg = f"{selection} : {message}"
        self.log.append(msg)
        self.events.set_text("\n".join(self.log))

    def destroy(self, *_args):
        Gtk.main_quit()

    def show_with_focus(self):
        force_focus()
        self.window.show_all()
        self.window.present()


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.log import enable_color
    from xpra.platform.gui import init, set_default_icon
    with program_context("Clipboard-Test", "Clipboard Test Tool"):
        enable_color()

        set_default_icon("clipboard.png")
        init()

        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("clipboard test window")

        w = ClipboardStateInfoWindow()
        GLib.idle_add(w.show_with_focus)
        Gtk.main()


if __name__ == "__main__":
    main()
    sys.exit(0)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2009 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

GTK_VERSION_INFO: dict[str, tuple] = {}


def get_gtk_version_info() -> dict[str, Any]:
    from xpra.util.version import parse_version

    # update props given:
    global GTK_VERSION_INFO
    if GTK_VERSION_INFO:
        return GTK_VERSION_INFO.copy()

    def av(k, v) -> None:
        GTK_VERSION_INFO[k] = parse_version(v)

    def V(k, module, attr_name) -> None:
        v = getattr(module, attr_name, None)
        if v is not None:
            av(k, v)

    from xpra.os_util import gi_import
    # this isn't the actual version, (only shows as "3.0")
    # but still better than nothing:
    import gi
    V("gi", gi, "__version__")

    def giv(k: str, gimod: str, attr_name: str) -> None:
        mod = gi_import(gimod)
        if mod:
            V(k, mod, attr_name)

    giv("gobject", "GObject", "pygobject_version")
    giv("gtk", "Gtk", "_version")
    giv("gdk", "Gdk", "_version")
    giv("gobject", "GObject", "_version")
    giv("pixbuf", "GdkPixbuf", "_version")
    giv("pixbuf", "GdkPixbuf", "PIXBUF_VERSION")

    def MAJORMICROMINOR(name: str, module) -> None:
        try:
            v = tuple(getattr(module, x) for x in ("MAJOR_VERSION", "MICRO_VERSION", "MINOR_VERSION"))
            av(name, ".".join(str(x) for x in v))
        except Exception:
            pass

    MAJORMICROMINOR("gtk", gi_import("Gtk"))
    MAJORMICROMINOR("glib", gi_import("GLib"))
    try:
        import cairo
        av("cairo", parse_version(cairo.version_info))  # pylint: disable=no-member
    except ImportError:
        pass
    try:
        pango = gi_import("Pango")
        av("pango", parse_version(pango.version_string()))
    except ImportError:
        pass
    return GTK_VERSION_INFO.copy()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform import platform_import

# default:
Keyboard: type | None = None

platform_import(globals(), "keyboard", True,
                "Keyboard")


def main():
    import sys
    from xpra.os_util import OSX, POSIX
    from xpra.util.str_fn import print_nested_dict
    from xpra.util.str_fn import csv
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("Keyboard-Tool", "Keyboard Tool"):
        # use the logger for the platform module we import from
        enable_color()
        verbose = consume_verbose_argv(sys.argv, "keyboard")

        # naughty, but how else can I hook this up?
        if POSIX and not OSX:
            try:
                from xpra.x11.bindings.posix_display_source import init_posix_display_source
                init_posix_display_source()
            except Exception as e:
                print("failed to connect to the X11 server:")
                print(" %s" % e)
                # hope for the best..

        if not Keyboard:
            print("no keyboard implementation")
            return 1
        keyboard = Keyboard()  # pylint: disable=not-callable
        mod_meanings, mod_managed, mod_pointermissing = keyboard.get_keymap_modifiers()
        print("Modifiers:")
        print_nested_dict(mod_meanings)
        print("")
        print("Server Managed                    : %s" % (csv(mod_managed) or "None"))
        print("Missing from pointer events       : %s" % (csv(mod_pointermissing) or "None"))
        print("")
        model, layout, layouts, variant, variants, options = keyboard.get_layout_spec()
        print(f"Model:      {model!r}")
        print(f"Layout:     {layout!r}")
        print(f"Layouts:    {csv(layouts)}")
        print(f"Variant:    {variant!r}")
        print(f"Variants:   {csv(variants)}")
        print(f"Options:    {options!r}")
        print("")
        print(f"Repeat:     {csv(keyboard.get_keyboard_repeat())}")
        if verbose and POSIX:
            keysyms = keyboard.get_x11_keymap()
            if keysyms:
                print("Keysyms:")
                for keycode, keysyms in keysyms.items():
                    print(" %3i    : %s" % (keycode, csv(keysyms)))
    return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

"""
Functions for converting to and from X11 properties.
    prop_encode
    prop_decode
"""

import struct
from io import BytesIO
from collections.abc import Callable, Sequence
from typing import Any, NoReturn, Final

from xpra.util.str_fn import hexstr
from xpra.x11.bindings.window import constants
from xpra.log import Logger

log = Logger("x11", "window")


USPosition: Final[int] = constants["USPosition"]
PPosition: Final[int] = constants["PPosition"]
PMaxSize: Final[int] = constants["PMaxSize"]
PMinSize: Final[int] = constants["PMinSize"]
PBaseSize: Final[int] = constants["PBaseSize"]
PResizeInc: Final[int] = constants["PResizeInc"]
PAspect: Final[int] = constants["PAspect"]
PWinGravity: Final[int] = constants["PWinGravity"]
XUrgencyHint: Final[int] = constants["XUrgencyHint"]
WindowGroupHint: Final[int] = constants["WindowGroupHint"]
StateHint: Final[int] = constants["StateHint"]
IconicState: Final[int] = constants["IconicState"]
InputHint: Final[int] = constants["InputHint"]

sizeof_long: Final[int] = struct.calcsize(b'@L')


def unsupported(*_args) -> NoReturn:
    raise RuntimeError("unsupported")


def _force_length(name: str, data: bytes, length: int, noerror_length: int = -1) -> bytes:
    if len(data) == length:
        return data
    if len(data) != noerror_length:
        log.warn("Odd-lengthed property %s: wanted %s bytes, got %s: %r"
                 % (name, length, len(data), data))
    # Zero-pad data
    data += b"\0" * length
    return data[:length]


class NetWMStrut:
    def __init__(self, data: bytes):
        # This eats both _NET_WM_STRUT and _NET_WM_STRUT_PARTIAL.  If we are
        # given a _NET_WM_STRUT instead of a _NET_WM_STRUT_PARTIAL, then it
        # will be only length 4 instead of 12, we just don't define the other values
        # and let the client deal with it appropriately
        if len(data) == 4 * sizeof_long:
            self.left, self.right, self.top, self.bottom = struct.unpack(b"@LLLL", data)
        else:

            data = _force_length("_NET_WM_STRUT or _NET_WM_STRUT_PARTIAL", data, 12 * sizeof_long)
            (
                self.left, self.right, self.top, self.bottom,
                self.left_start_y, self.left_end_y,
                self.right_start_y, self.right_end_y,
                self.top_start_x, self.top_end_x,
                self.bottom_start_x, self.bottom_stop_x,
            ) = struct.unpack(b"@" + b"L" * 12, data)

    def todict(self):
        return self.__dict__

    def __str__(self):
        return "NetWMStrut(%s)" % self.todict()


class MotifWMHints:
    __slots__ = ("flags", "functions", "decorations", "input_mode", "status")

    def __init__(self, data: bytes):
        # some applications use the wrong size (ie: blender uses 16) so pad it:
        pdata = _force_length("_MOTIF_WM_HINTS", data, sizeof_long*5, sizeof_long*4)
        self.flags, self.functions, self.decorations, self.input_mode, self.status = \
            struct.unpack(b"@LLLlL", pdata)
        log("MotifWMHints(%s)=%s", hexstr(data), self)

    # found in mwmh.h:
    # "flags":
    FUNCTIONS_BIT   = 0
    DECORATIONS_BIT = 1
    INPUT_MODE_BIT  = 2
    STATUS_BIT      = 3
    # "functions":
    ALL_BIT         = 0
    RESIZE_BIT      = 1
    MOVE_BIT        = 2      # like _NET_WM_ACTION_MOVE
    MINIMIZE_BIT    = 3      # like _NET_WM_ACTION_MINIMIZE
    MAXIMIZE_BIT    = 4      # like _NET_WM_ACTION_(FULLSCREEN|MAXIMIZE_(HORZ|VERT))
    CLOSE_BIT       = 5      # like _NET_WM_ACTION_CLOSE
    SHADE_BIT       = 6      # like _NET_WM_ACTION_SHADE
    STICK_BIT       = 7      # like _NET_WM_ACTION_STICK
    FULLSCREEN_BIT  = 8      # like _NET_WM_ACTION_FULLSCREEN
    ABOVE_BIT       = 9      # like _NET_WM_ACTION_ABOVE
    BELOW_BIT       = 10     # like _NET_WM_ACTION_BELOW
    MAXIMUS_BIT     = 11     # like _NET_WM_ACTION_MAXIMUS_(LEFT|RIGHT|TOP|BOTTOM)
    # "decorations":
    # ALL_BIT         = 0 (same as above)
    BORDER_BIT      = 1
    RESIZEH_BIT     = 2
    TITLE_BIT       = 3
    MENU_BIT        = 4
    # MINIMIZE_BIT    = 5 (same as above)
    # MAXIMIZE_BIT    = 6 (same as above)
    # CLOSE_BIT                # non-standard close button
    # RESIZE_BIT               # non-standard resize button
    # SHADE_BIT,               # non-standard shade button
    # STICK_BIT,               # non-standard stick button
    # MAXIMUS_BIT              # non-standard maxim
    # "input":
    MODELESS        = 0
    PRIMARY_APPLICATION_MODAL = 1
    SYSTEM_MODAL    = 2
    FULL_APPLICATION_MODAL = 3
    # "status":
    TEAROFF_WINDOW  = 0

    FLAGS_STR = {
        FUNCTIONS_BIT      : "functions",
        DECORATIONS_BIT    : "decorations",
        INPUT_MODE_BIT     : "input",
        STATUS_BIT         : "status",
    }
    FUNCTIONS_STR = {
        ALL_BIT        : "all",
        RESIZE_BIT     : "resize",
        MOVE_BIT       : "move",
        MINIMIZE_BIT   : "minimize",
        MAXIMIZE_BIT   : "maximize",
        CLOSE_BIT      : "close",
        SHADE_BIT      : "shade",
        STICK_BIT      : "stick",
        FULLSCREEN_BIT : "fullscreen",
        ABOVE_BIT      : "above",
        BELOW_BIT      : "below",
        MAXIMUS_BIT    : "maximus",
    }
    DECORATIONS_STR = {
        ALL_BIT      : "all",
        BORDER_BIT   : "border",
        RESIZEH_BIT  : "resizeh",
        TITLE_BIT    : "title",
        MENU_BIT     : "menu",
        MINIMIZE_BIT : "minimize",
        MAXIMIZE_BIT : "maximize",
    }
    INPUT_STR = {
        MODELESS                   : "modeless",
        PRIMARY_APPLICATION_MODAL  : "primary-application-modal",
        SYSTEM_MODAL               : "system-modal",
        FULL_APPLICATION_MODAL     : "full-application-modal",
    }

    STATUS_STR = {
        TEAROFF_WINDOW : "tearoff",
    }

    def bits_to_strs(self, int_val: int, flag_bit: int, dict_str: dict) -> Sequence[str]:
        if flag_bit and not self.flags & (2**flag_bit):
            # the bit is not set, ignore this attribute
            return ()
        return tuple(v for k, v in dict_str.items() if int_val & (2**k))

    def flags_strs(self) -> Sequence[str]:
        return self.bits_to_strs(self.flags,
                                 0,
                                 MotifWMHints.FLAGS_STR)

    def functions_strs(self) -> Sequence[str]:
        return self.bits_to_strs(self.functions,
                                 MotifWMHints.FUNCTIONS_BIT,
                                 MotifWMHints.FUNCTIONS_STR)

    def decorations_strs(self) -> Sequence[str]:
        return self.bits_to_strs(self.decorations,
                                 MotifWMHints.DECORATIONS_BIT,
                                 MotifWMHints.DECORATIONS_STR)

    def input_strs(self) -> str:
        if self.flags & (2**MotifWMHints.INPUT_MODE_BIT):
            return MotifWMHints.INPUT_STR.get(self.input_mode, "unknown mode: %i" % self.input_mode)
        return "modeless"

    def status_strs(self) -> Sequence[str]:
        return self.bits_to_strs(self.input_mode,
                                 MotifWMHints.STATUS_BIT,
                                 MotifWMHints.STATUS_STR)

    def __str__(self):
        attrs = {
            "flags": self.flags_strs(),
            "functions": self.functions_strs(),
            "decorations": self.decorations_strs(),
            "input_mode": self.input_strs(),
            "status": self.status_strs(),
        }
        return f"MotifWMHints({attrs})"


def _read_image(stream) -> tuple[int, int, str, bytes] | None:
    try:
        int_size = struct.calcsize(b"@I")
        long_size = struct.calcsize(b"@L")
        header = stream.read(long_size*2)
        if not header:
            return None
        width, height = struct.unpack(b"@LL", header)
        data = stream.read(width * height * long_size)
        expected = width * height * long_size
        if len(data) < expected:
            log.warn("Warning: corrupt _NET_WM_ICON, expected %i bytes but got %i",
                     expected, len(data))
            return None
        if int_size != long_size:
            # long to ints (CARD32):
            longs = struct.unpack(b"@"+b"l"*(width*height), data)
            data = struct.pack(b"@"+b"i"*(width*height), *longs)
    except Exception:
        log.warn("Weird corruption in _NET_WM_ICON", exc_info=True)
        return None
    return width, height, "BGRA", data

# This returns a list of icons from a _NET_WM_ICON property.
# each icon is a tuple:
# (width, height, fmt, data)


def NetWMIcons(data) -> list | None:
    icons = []
    stream = BytesIO(data)
    while True:
        icon = _read_image(stream)
        if icon is None:
            break
        icons.append(icon)
    if not icons:
        return None
    return icons


def _to_latin1(v: str) -> bytes:
    return v.encode("latin1")


def _from_latin1(v: bytes) -> str:
    return v.decode("latin1")


def _to_utf8(v: str) -> bytes:
    return v.encode("UTF-8")


def _from_utf8(v: bytes) -> str:
    return v.decode("UTF-8")


def _to_long(v: int) -> bytes:
    return struct.pack(b"@L", v)


def _from_long(v: bytes) -> int:
    return struct.unpack(b"@L", v)[0]


def _to_state(v: int) -> bytes:
    return struct.pack(b"@LL", v, 0)


def _from_state(v: bytes) -> int:
    wm_state = _force_length("WM_STATE", v, sizeof_long*2, sizeof_long)
    return struct.unpack(b"@LL", wm_state)[0]


PROP_TYPES: dict[str, tuple[type, str, int, Callable[[Any], bytes], Callable[[bytes], Any], bytes | None]] = {
    # Python type, X type Atom, formatbits, serializer, deserializer, list terminator
    "utf8": (str, "UTF8_STRING", 8, _to_utf8, _from_utf8, b"\0"),
    # In theory, there should be something clever about COMPOUND_TEXT here.  I
    # am not sufficiently clever to deal with COMPOUNT_TEXT.  Even knowing
    # that Xutf8TextPropertyToTextlist exists.
    "latin1": (str, "STRING", 8, _to_latin1, _from_latin1, b"\0"),
    "state": (int, "WM_STATE", 32, _to_state, _from_state, b""),
    "u32": (int, "CARDINAL", 32, _to_long, _from_long, b""),
    "integer": (int, "INTEGER", 32, _to_long, _from_long, b""),
    "strut": (NetWMStrut, "CARDINAL", 32, unsupported, NetWMStrut, None),
    "strut-partial": (NetWMStrut, "CARDINAL", 32, unsupported, NetWMStrut, None),
    "motif-hints": (MotifWMHints, "_MOTIF_WM_HINTS", 32, unsupported, MotifWMHints, None),
    "icons": (list, "CARDINAL", 32, unsupported, NetWMIcons, None),
    "window": (int, "WINDOW", 32, _to_long, _from_long, b""),
}

PROP_SIZES = {
    "icons": 4*1024*1024,
}


def prop_encode(etype : list | tuple | str, value):
    if isinstance(etype, (list, tuple)):
        return _prop_encode_list(etype[0], value)
    return _prop_encode_scalar(etype, value)


def _prop_encode_scalar(etype: str, value) -> tuple[str, int, bytes]:
    pytype, atom, formatbits, serialize = PROP_TYPES[etype][:4]
    assert isinstance(value, pytype), "value for atom %s is not a %s: %s" % (atom, pytype, type(value))
    return atom, formatbits, serialize(value)


def _prop_encode_list(etype: str, value) -> tuple[str, int, bytes]:
    _, atom, formatbits, _, _, terminator = PROP_TYPES[etype]
    if terminator is None:
        raise ValueError(f"cannot encode lists of {etype!r}")
    value = tuple(value)
    serialized = tuple(_prop_encode_scalar(etype, v)[2] for v in value)
    # Strings in X really are null-separated, not null-terminated (ICCCM
    # 2.7.1, see also note in 4.1.2.5)
    return atom, formatbits, terminator.join(x for x in serialized if x is not None)


def prop_decode(etype: str | list | tuple, data: bytes):
    if isinstance(etype, (list, tuple)):
        return _prop_decode_list(etype[0], data)
    return _prop_decode_scalar(etype, data)


def _prop_decode_scalar(etype: str, data: bytes):
    pytype, _, _, _, deserialize, _ = PROP_TYPES[etype]
    value = deserialize(data)
    assert value is None or isinstance(value, pytype), "expected a %s but value is a %s" % (pytype, type(value))
    return value


def _prop_decode_list(etype: str, data) -> list:
    _, _, formatbits, _, _, terminator = PROP_TYPES[etype]
    if terminator:
        datums = data.split(terminator)
    else:
        datums = []
        if formatbits == 32:
            nbytes = struct.calcsize("@L")
        else:
            nbytes = formatbits // 8
        while data:
            datums.append(data[:nbytes])
            data = data[nbytes:]
    props = (_prop_decode_scalar(etype, datum) for datum in datums)
    return [x for x in props if x is not None]



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2009 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=no-member

""" launcher.py

This was meant to be a simple GUI for starting the xpra client.
Then it morphed into something quite messy. Sorry.

"""

import os.path
import sys
import traceback
from typing import Any
from collections.abc import Callable

from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
from xpra.gtk.signals import register_os_signals
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import (
    scaled_image, imagebutton, label, choose_file, modify_fg,
    set_widget_bg_color, set_widget_fg_color, red, black,
)
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.util.str_fn import csv, repr_ellipsized
from xpra.os_util import WIN32, OSX, gi_import
from xpra.util.env import IgnoreWarningsContext
from xpra.net.common import DEFAULT_PORT
from xpra.util.thread import start_thread
from xpra.gtk.dialogs.about import about
from xpra.gtk.dialogs.util import hb_button
from xpra.scripts.main import (
    connect_to, make_client, bypass_no_gtk,
    configure_network, configure_env, configure_logging,
)
from xpra.common import noop
from xpra.scripts.parsing import is_local, get_ssh_args, parse_ssh_option, get_ssh_proxy_args, get_ssl_options
from xpra.exit_codes import RETRY_EXIT_CODES, ExitCode, ExitValue, exit_str
from xpra.platform.info import get_username
from xpra.log import Logger, enable_debug_for

log = Logger("launcher")

Gtk = gi_import("Gtk")
GdkPixbuf = gi_import("GdkPixbuf")
GLib = gi_import("GLib")

MODE_NESTED_SSH = "ssh -> ssh"
MODE_SSH = "ssh"
MODE_TCP = "tcp"
MODE_SSL = "ssl"
MODE_WS = "ws"
MODE_WSS = "wss"
MODE_QUIC = "quic"

# what we save in the config file:
SAVED_FIELDS = [
    "username", "password", "host", "port", "mode", "ssh_port", "path",
    "encoding", "quality", "min-quality", "speed", "min-speed",
    "proxy_port", "proxy_username", "proxy_key", "proxy_password",
    "proxy_host",
    "autoconnect",
]

# options not normally found in xpra config file
# but which can be present in a launcher config:
LAUNCHER_OPTION_TYPES = {
    "host": str,
    "port": int,
    "username": str,
    "password": str,
    "mode": str,
    "path": str,
    "autoconnect": bool,
    "ssh_port": int,
    "proxy_host": str,
    "proxy_port": int,
    "proxy_username": str,
    "proxy_password": str,
    "proxy_key": str,
}
LAUNCHER_DEFAULTS = {
    "host": "",
    "port": -1,
    "username": get_username(),
    "password": "",
    "mode": MODE_TCP,  # tcp,ssh,..
    "path": "",
    "autoconnect": False,
    "ssh_port": 22,
    "proxy_host": "",
    "proxy_port": 22,
    "proxy_username": get_username(),
    "proxy_password": "",
    "proxy_key": "",
}


def get_active_item_index(optionmenu) -> int:
    menu = optionmenu.get_popup()
    for i, x in enumerate(menu.get_children()):
        if hasattr(x, "get_active") and x.get_active():
            return i
    return -1


def set_history_from_active(optionmenu) -> None:
    # Used for MenuButton combo:
    # sets the first active menu entry as the "history" value (the selected item)
    i = get_active_item_index(optionmenu)
    if i > 0:
        optionmenu.set_history(i)


def has_mdns() -> bool:
    try:
        # pylint: disable=import-outside-toplevel
        from xpra.net.mdns import get_listener_class
        lc = get_listener_class()
        log("mdns listener class: %s", lc)
        if lc:
            return True
    except ImportError as e:
        log("no mdns support: %s", e)
    return False


def get_connection_modes() -> list[str]:
    modes = [MODE_SSH, MODE_NESTED_SSH]
    try:
        import ssl
        assert ssl
        modes.append(MODE_SSL)
        modes.append(MODE_QUIC)
    except ImportError:
        pass
    # assume crypto is available
    try:
        from xpra.net.crypto import get_modes
        for mode in get_modes():
            modes.append(f"tcp + aes-{mode.lower()}")
    except ImportError:
        pass
    modes.append(MODE_TCP)
    modes.append(MODE_WS)
    modes.append(MODE_WSS)
    return modes


def image_button(text="", tooltip="", icon_pixbuf=None, clicked_cb=None) -> Gtk.Image:
    icon = Gtk.Image()
    icon.set_from_pixbuf(icon_pixbuf)
    return imagebutton(text, icon, tooltip, clicked_cb, icon_size=None)


def button(tooltip, icon_name, callback) -> Gtk.Button:
    btn = Gtk.Button()
    theme = Gtk.IconTheme.get_default()
    try:
        pixbuf = theme.load_icon(icon_name, Gtk.IconSize.BUTTON, Gtk.IconLookupFlags.USE_BUILTIN)
    except Exception:
        pixbuf = None
    if not pixbuf:
        pixbuf = get_icon_pixbuf(f"{icon_name}.png")
        if pixbuf:
            with IgnoreWarningsContext():
                for size in (16, 32, 48):
                    scaled = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)
                    Gtk.IconTheme.add_builtin_icon(icon_name, size, scaled)
            try:
                builtin_reload = theme.load_icon(icon_name, Gtk.IconSize.BUTTON, Gtk.IconLookupFlags.USE_BUILTIN)
            except Exception:
                log("button: failed to load icon after adding to builtins", exc_info=True)
                size = 32
                pixbuf = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)
            else:
                pixbuf = builtin_reload
    if pixbuf:
        image = Gtk.Image.new_from_pixbuf(pixbuf)
        btn.add(image)
    btn.set_tooltip_text(tooltip)

    def clicked(*_args) -> None:
        callback()

    btn.connect("clicked", clicked)
    return btn


def get_launcher_validation() -> dict:
    # TODO: since "mode" is not part of global options
    # this validation should be injected from the launcher instead
    def validate_in_list(x, options) -> str:
        if x in options:
            return ""
        return "must be in " + csv(options)

    modes = get_connection_modes()
    return {"mode": lambda x: validate_in_list(x, modes)}


def accel_close(*args) -> None:
    log("accel_close%s", args)
    Gtk.main_quit()


class ApplicationWindow:

    def __init__(self):
        # Default connection options
        self.config = make_defaults_struct(extras_defaults=LAUNCHER_DEFAULTS,
                                           extras_types=LAUNCHER_OPTION_TYPES,
                                           extras_validation=get_launcher_validation())
        self.parse_ssh()
        # TODO: the fixup does not belong here?
        from xpra.scripts.main import fixup_options
        fixup_options(self.config)
        # what we save by default:
        self.config_keys = set(SAVED_FIELDS)
        self.client = None
        self.exit_launcher = False
        self.exit_code = None
        self.current_error = None

    def parse_ssh(self) -> None:
        ssh_cmd = parse_ssh_option(self.config.ssh)[0].strip().lower()
        self.is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
        self.is_paramiko = ssh_cmd.startswith("paramiko")

    def create_window_with_config(self) -> None:
        self.do_create_window()
        self.update_gui_from_config()

    def do_create_window(self) -> None:
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.destroy)
        self.window.set_default_size(400, 260)
        self.window.set_title("Xpra Launcher")
        self.window.set_position(Gtk.WindowPosition.CENTER)
        with IgnoreWarningsContext():
            self.window.set_wmclass("xpra-launcher-gui", "Xpra-Launcher-GUI")
        add_close_accel(self.window, accel_close)
        icon = get_icon_pixbuf("connect.png")
        if icon:
            self.window.set_icon(icon)

        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = "Session Launcher"
        self.window.set_titlebar(hb)

        def show_about(*_args) -> None:
            about(parent=self.window)

        hb.add(hb_button("About", "help-about", show_about))
        self.bug_tool = None

        def bug(*_args) -> None:
            if self.bug_tool is None:
                from xpra.gtk.dialogs.bug_report import BugReport
                self.bug_tool = BugReport()
                self.bug_tool.init(show_about=False)
            self.bug_tool.show()

        hb.add(hb_button("Bug Report", "bugs", bug))
        if has_mdns():
            self.mdns_gui = None

            def mdns(*_args) -> None:
                if self.mdns_gui is None:
                    from xpra.gtk.dialogs.mdns_gui import mdns_sessions
                    self.mdns_gui = mdns_sessions(self.config)

                    def close_mdns() -> None:
                        mdns_gui = self.mdns_gui
                        if mdns_gui:
                            self.mdns_gui = None
                            mdns_gui.close()

                    self.mdns_gui.do_quit = close_mdns
                else:
                    self.mdns_gui.present()

            hb.add(hb_button("Browse Sessions", "mdns", mdns))
        hb.show_all()

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(15)

        # Title
        vbox.pack_start(label("Connect to xpra server", font="sans 14"))

        # Mode:
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.mode_combo = Gtk.ComboBoxText()
        for x in get_connection_modes():
            self.mode_combo.append_text(x.upper())
        self.mode_combo.connect("changed", self.mode_changed)
        hbox.pack_start(label("Mode: "), False, False)
        hbox.pack_start(self.mode_combo, False, False)
        align_hbox = Gtk.Alignment(xalign=0.5)
        align_hbox.add(hbox)
        vbox.pack_start(align_hbox)

        # Username@Host:Port (ssh -> ssh, proxy)
        vbox_proxy = Gtk.VBox(homogeneous=False, spacing=15)
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.proxy_vbox = vbox_proxy
        self.proxy_username_entry = Gtk.Entry()
        self.proxy_username_entry.set_max_length(128)
        self.proxy_username_entry.set_width_chars(16)
        self.proxy_username_entry.connect("changed", self.validate)
        self.proxy_username_entry.connect("activate", self.connect_clicked)
        self.proxy_username_entry.set_tooltip_text("username")
        self.proxy_host_entry = Gtk.Entry()
        self.proxy_host_entry.set_max_length(128)
        self.proxy_host_entry.set_width_chars(24)
        self.proxy_host_entry.connect("changed", self.validate)
        self.proxy_host_entry.connect("activate", self.connect_clicked)
        self.proxy_host_entry.set_tooltip_text("hostname")
        self.proxy_port_entry = Gtk.Entry()
        self.proxy_port_entry.set_max_length(5)
        self.proxy_port_entry.set_width_chars(5)
        self.proxy_port_entry.connect("changed", self.validate)
        self.proxy_port_entry.connect("activate", self.connect_clicked)
        self.proxy_port_entry.set_tooltip_text("SSH port")
        hbox.pack_start(label("Proxy: "), False, False)
        hbox.pack_start(self.proxy_username_entry, True, True)
        hbox.pack_start(label("@"), False, False)
        hbox.pack_start(self.proxy_host_entry, True, True)
        hbox.pack_start(self.proxy_port_entry, False, False)
        vbox_proxy.pack_start(hbox)

        # Password
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.proxy_password_hbox = hbox
        self.proxy_password_entry = Gtk.Entry()
        self.proxy_password_entry.set_max_length(128)
        self.proxy_password_entry.set_width_chars(30)
        self.proxy_password_entry.set_text("")
        self.proxy_password_entry.set_visibility(False)
        self.proxy_password_entry.connect("changed", self.password_ok)
        self.proxy_password_entry.connect("changed", self.validate)
        self.proxy_password_entry.connect("activate", self.connect_clicked)
        hbox.pack_start(label("Proxy Password"), False, False)
        hbox.pack_start(self.proxy_password_entry, True, True)
        vbox_proxy.pack_start(hbox)

        # Private key
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.pkey_hbox = hbox
        self.proxy_key_label = label("Proxy private key path (PPK):")
        self.proxy_key_entry = Gtk.Entry()
        self.proxy_key_browse = Gtk.Button(label="Browse")
        self.proxy_key_browse.connect("clicked", self.proxy_key_browse_clicked)
        hbox.pack_start(self.proxy_key_label, False, False)
        hbox.pack_start(self.proxy_key_entry, True, True)
        hbox.pack_start(self.proxy_key_browse, False, False)
        vbox_proxy.pack_start(hbox)

        # Check boxes
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.check_boxes_hbox = hbox
        self.password_scb = Gtk.CheckButton(label="Server password same as proxy")
        self.password_scb.set_mode(True)
        self.password_scb.set_active(True)
        self.password_scb.connect("toggled", self.validate)
        align_password_scb = Gtk.Alignment(xalign=1.0)
        align_password_scb.add(self.password_scb)
        self.username_scb = Gtk.CheckButton(label="Server username same as proxy")
        self.username_scb.set_mode(True)
        self.username_scb.set_active(True)
        self.username_scb.connect("toggled", self.validate)
        align_username_scb = Gtk.Alignment(xalign=0.0)
        align_username_scb.add(self.username_scb)
        hbox.pack_start(align_username_scb, True, True)
        hbox.pack_start(align_password_scb, True, True)
        vbox_proxy.pack_start(hbox)

        # condiditonal stuff that goes away for "normal" ssh
        vbox.pack_start(vbox_proxy)

        # Username@Host:Port (main)
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.username_entry = Gtk.Entry()
        self.username_entry.set_max_length(128)
        self.username_entry.set_width_chars(16)
        self.username_entry.connect("changed", self.validate)
        self.username_entry.connect("activate", self.connect_clicked)
        self.username_entry.set_tooltip_text("server username")
        self.host_entry = Gtk.Entry()
        self.host_entry.set_max_length(128)
        self.host_entry.set_width_chars(24)
        self.host_entry.connect("changed", self.validate)
        self.host_entry.connect("activate", self.connect_clicked)
        self.host_entry.set_tooltip_text("server hostname or IP address")
        self.ssh_port_entry = Gtk.Entry()
        self.ssh_port_entry.set_max_length(5)
        self.ssh_port_entry.set_width_chars(5)
        self.ssh_port_entry.connect("changed", self.validate)
        self.ssh_port_entry.connect("activate", self.connect_clicked)
        self.ssh_port_entry.set_tooltip_text("SSH port")
        self.port_entry = Gtk.Entry()
        self.port_entry.set_max_length(5)
        self.port_entry.set_width_chars(5)
        self.port_entry.connect("changed", self.validate)
        self.port_entry.connect("activate", self.connect_clicked)
        self.port_entry.set_tooltip_text("port/display")
        self.path_entry = Gtk.Entry()
        self.path_entry.set_max_length(128)
        self.path_entry.set_width_chars(5)
        self.path_entry.connect("activate", self.connect_clicked)
        self.path_entry.set_tooltip_text("request path")
        hbox.pack_start(label("Server:"), False, False)
        hbox.pack_start(self.username_entry, True, True)
        hbox.pack_start(label("@"), False, False)
        hbox.pack_start(self.host_entry, True, True)
        hbox.pack_start(self.ssh_port_entry, False, False)
        hbox.pack_start(label(":"), False, False)
        hbox.pack_start(self.port_entry, False, False)
        hbox.pack_start(label("/"), False, False)
        hbox.pack_start(self.path_entry, False, False)
        vbox.pack_start(hbox)

        # Password
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.password_hbox = hbox
        self.password_entry = Gtk.Entry()
        self.password_entry.set_max_length(128)
        self.password_entry.set_width_chars(30)
        self.password_entry.set_text("")
        self.password_entry.set_visibility(False)
        self.password_entry.connect("changed", self.password_ok)
        self.password_entry.connect("changed", self.validate)
        self.password_entry.connect("activate", self.connect_clicked)
        hbox.pack_start(label("Server Password:"), False, False)
        hbox.pack_start(self.password_entry, True, True)
        vbox.pack_start(hbox)

        # strict host key check for SSL and SSH
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.nostrict_host_check = Gtk.CheckButton(label="Disable Strict Host Key Check")
        self.nostrict_host_check.set_active(False)
        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
        al.add(self.nostrict_host_check)
        hbox.pack_start(al)
        vbox.pack_start(hbox)

        # auto-connect
        hbox = Gtk.HBox(homogeneous=False, spacing=5)
        self.autoconnect = Gtk.CheckButton(label="Auto-connect")
        self.autoconnect.set_active(False)
        self.autoconnect.set_tooltip_text("Connect without opening this launcher when opening the session file")
        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
        al.add(self.autoconnect)
        hbox.pack_start(al)
        vbox.pack_start(hbox)

        # Info Label
        self.info = label()
        self.info.set_line_wrap(True)
        self.info.set_size_request(360, -1)
        modify_fg(self.info, red)
        vbox.pack_start(self.info)

        # Buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox)
        # Save:
        self.save_btn = Gtk.Button(label="Save")
        self.save_btn.set_tooltip_text("Save settings to a session file")
        self.save_btn.connect("clicked", self.save_clicked)
        hbox.pack_start(self.save_btn)
        # Load:
        self.load_btn = Gtk.Button(label="Load")
        self.load_btn.set_tooltip_text("Load settings from a session file")
        self.load_btn.connect("clicked", self.load_clicked)
        hbox.pack_start(self.load_btn)
        # Connect button:
        self.connect_btn = Gtk.Button(label="Connect")
        self.connect_btn.connect("clicked", self.connect_clicked)
        connect_icon = get_icon_pixbuf("retry.png")
        if connect_icon:
            self.connect_btn.set_image(scaled_image(connect_icon, 24))
        hbox.pack_start(self.connect_btn)

        vbox.show_all()
        self.window.vbox = vbox
        self.window.add(vbox)

    def validate(self, *args) -> list:
        mode = self.mode_combo.get_active_text().lower()
        ssh = mode == MODE_SSH
        sshtossh = mode == MODE_NESTED_SSH
        errs = []
        host = self.host_entry.get_text()
        errs.append((self.host_entry, not bool(host), "specify the host"))
        if ssh or sshtossh:
            # validate ssh port:
            ssh_port = self.ssh_port_entry.get_text()
            try:
                ssh_port = int(ssh_port)
            except ValueError:
                ssh_port = -1
            errs.append((self.ssh_port_entry,
                         ssh_port < 0 or ssh_port >= 2 ** 16,
                         "invalid SSH port number"))
        if sshtossh:
            # validate ssh port:
            proxy_port = self.proxy_port_entry.get_text()
            try:
                proxy_port = int(proxy_port)
            except ValueError:
                proxy_port = -1
            errs.append((self.proxy_port_entry,
                         proxy_port < 0 or proxy_port >= 2 ** 16,
                         "invalid SSH port number"))
        port = self.port_entry.get_text()
        if sshtossh:
            if self.password_scb.get_active():
                self.password_entry.set_sensitive(False)
                self.password_entry.set_text(self.proxy_password_entry.get_text())
            else:
                self.password_entry.set_sensitive(True)
            if self.username_scb.get_active():
                self.username_entry.set_sensitive(False)
                self.username_entry.set_text(self.proxy_username_entry.get_text())
            else:
                self.username_entry.set_sensitive(True)
            errs.append((self.proxy_host_entry,
                         not bool(self.proxy_host_entry.get_text()),
                         "specify the proxy host"))
        # check username *after* the checkbox action
        if ssh or sshtossh:
            errs.append((self.username_entry,
                         not bool(self.username_entry.get_text()),
                         "specify username"))
        if sshtossh:
            errs.append((self.proxy_username_entry,
                         not bool(self.proxy_username_entry.get_text()),
                         "specify proxy username"))
        if ssh or sshtossh and not port:
            port = 0  # port optional with ssh
        else:
            try:
                port = int(port)
            except (ValueError, TypeError):
                port = -1
        errs.append((self.port_entry, port < 0 or port >= 2 ** 16, "invalid port number"))
        err_text = []
        for w, e, text in errs:
            set_widget_bg_color(w, e)
            if e:
                err_text.append(text)
        log(f"validate({args}) err_text={err_text}, errs={errs}")
        self.set_info_text(csv(err_text), bool(err_text))
        self.connect_btn.set_sensitive(len(err_text) == 0)
        return errs

    def show(self) -> None:
        self.window.show()
        self.window.present()
        self.connect_btn.grab_focus()

    def run(self) -> ExitValue:
        Gtk.main()
        return self.exit_code or 0

    def mode_changed(self, *_args) -> None:
        mode = self.mode_combo.get_active_text().lower()
        ssh = mode == MODE_SSH
        sshtossh = mode == MODE_NESTED_SSH
        if ssh or sshtossh:
            self.port_entry.set_tooltip_text("Display number (optional)")
            self.port_entry.set_text("")
            self.ssh_port_entry.show()
            self.password_entry.set_tooltip_text("SSH Password")
            self.username_entry.set_tooltip_text("SSH Username")
            if ssh:
                self.proxy_vbox.hide()
                self.password_scb.hide()
                self.password_entry.set_sensitive(True)
                self.username_entry.set_sensitive(True)
            if sshtossh:
                self.proxy_vbox.show()
                self.password_scb.show()
        else:
            self.password_entry.set_sensitive(True)
            self.username_entry.set_sensitive(True)
            self.proxy_vbox.hide()
            self.ssh_port_entry.hide()
            self.ssh_port_entry.set_text("")
            port_str = self.port_entry.get_text()
            if not port_str:
                self.port_entry.set_text(str(max(0, self.config.port) or DEFAULT_PORT))
            self.port_entry.set_tooltip_text("xpra server port number")
            self.password_entry.set_tooltip_text("Session Password (optional)")
            self.username_entry.set_tooltip_text("Session Username (optional)")
            if self.config.port > 0:
                self.port_entry.set_text(str(self.config.port))
        can_use_password = True
        sshpass = False
        if ssh or sshtossh:
            if not self.is_putty:
                self.proxy_key_entry.set_text("OpenSSH/Paramiko use ~/.ssh")
                self.proxy_key_entry.set_editable(False)
                self.proxy_key_entry.set_sensitive(False)
                self.proxy_key_browse.hide()
            if self.is_paramiko or self.is_putty:
                can_use_password = True
            else:
                # we can also use password if sshpass is installed:
                from xpra.platform.paths import get_sshpass_command
                sshpass = get_sshpass_command()
                can_use_password = bool(sshpass)
                sshpass = bool(sshpass)
        if can_use_password:
            self.password_hbox.show()
            if sshtossh:
                self.proxy_password_hbox.show()
                # sshpass cannot do different username/passwords for proxy and destination
                if not sshpass:
                    self.check_boxes_hbox.show()
                    p = self.password_entry.get_text()
                    pp = self.proxy_password_entry.get_text()
                    u = self.username_entry.get_text()
                    pu = self.proxy_username_entry.get_text()
                else:
                    self.check_boxes_hbox.hide()
                    p = pp = None
                    u = pu = None
                self.password_scb.set_active(p == pp)
                self.username_scb.set_active(u == pu)
        else:
            self.password_hbox.hide()
            if sshtossh:
                self.check_boxes_hbox.hide()
                self.proxy_password_hbox.hide()
        can_use_path = mode in ("ws", "wss")
        if can_use_path:
            self.path_entry.show()
        else:
            self.path_entry.hide()
        self.validate()
        if mode in (MODE_SSL, MODE_WSS, MODE_QUIC) or (mode == MODE_SSH and not WIN32):
            self.nostrict_host_check.show()
        else:
            self.nostrict_host_check.hide()

    def reset_errors(self) -> None:
        self.set_sensitive(True)
        self.set_info_text("")
        for widget in (
                self.info, self.password_entry, self.username_entry, self.host_entry,
                self.port_entry, self.proxy_password_entry, self.proxy_username_entry,
                self.proxy_host_entry, self.proxy_port_entry, self.ssh_port_entry,
        ):
            set_widget_fg_color(self.info, False)
            set_widget_bg_color(widget, False)

    def set_info_text(self, text, is_error=False) -> None:
        if self.info:
            def do_set_info() -> None:
                self.info.set_text(text)
                self.info.set_selectable(is_error)
                set_widget_fg_color(self.info, is_error)

            GLib.idle_add(do_set_info)

    def set_sensitive(self, s) -> None:
        GLib.idle_add(self.window.set_sensitive, s)

    def choose_pkey_file(self, title, action, action_button, callback) -> None:
        file_filter = Gtk.FileFilter()
        file_filter.set_name("All Files")
        file_filter.add_pattern("*")
        choose_file(self.window, title, action, action_button, callback, file_filter)

    def proxy_key_browse_clicked(self, *args) -> None:
        log("proxy_key_browse_clicked%s", args)

        def do_choose(filename) -> None:
            # make sure the file extension is .ppk
            if os.path.splitext(filename)[-1] != ".ppk":
                filename += ".ppk"
            self.proxy_key_entry.set_text(filename)

        self.choose_pkey_file("Choose SSH private key", Gtk.FileChooserAction.OPEN, Gtk.STOCK_OPEN, do_choose)

    def connect_clicked(self, *args) -> None:
        log(f"connect_clicked({args})")
        self.update_options_from_gui()
        self.do_connect()

    def clean_client(self) -> None:
        c = self.client
        if c:
            c.disconnect_and_quit = noop
            c.warn_and_quit = noop
            c.quit = noop
            c.exit = noop
            c.cleanup()

    def handle_exception(self, e) -> None:
        log(f"handle_exception({e})")
        t = str(e)
        log("handle_exception: %s", traceback.format_exc())
        if self.config.debug:
            # in debug mode, include the full stacktrace:
            t = traceback.format_exc()

        def ui_handle_exception() -> None:
            self.clean_client()
            self.set_sensitive(True)
            if not self.current_error:
                self.current_error = t
                self.set_info_text(t, True)
            self.window.show()
            self.window.present()

        GLib.idle_add(ui_handle_exception)

    def do_connect(self) -> None:
        try:
            self.connect_builtin()
        except Exception as e:
            log.error("Error: cannot connect", exc_info=True)
            self.handle_exception(e)

    def connect_builtin(self) -> None:
        # cooked vars used by connect_to
        username = self.config.username
        params = {
            "type": self.config.mode,
            "username": username,
        }
        if self.config.mode in (MODE_SSH, MODE_NESTED_SSH):
            if self.config.socket_dir:
                params["socket_dir"] = self.config.socket_dir
            params["remote_xpra"] = self.config.remote_xpra
            params["proxy_command"] = ["_proxy"]
            if self.config.port >= 0:
                params["display"] = f":{self.config.port}"
                params["display_as_args"] = [params["display"]]
            else:
                params["display"] = "auto"
                params["display_as_args"] = []
            params["ssh"] = self.config.ssh
            params["is_putty"] = self.is_putty
            params["is_paramiko"] = self.is_paramiko
            password = self.config.password
            host = self.config.host
            upos = host.find("@")
            if upos >= 0:
                # found at sign: username@host
                username = host[:upos]
                host = host[upos + 1:]
                ppos = username.find(":")
                if ppos >= 0:
                    # found separator: username:password@host
                    password = username[ppos + 1:]
                    username = username[:ppos]
            if self.config.ssh_port and self.config.ssh_port != 22:
                params["port"] = self.config.ssh_port
            params["host"] = host
            ssh_cmd = parse_ssh_option(self.config.ssh)
            ssh_cmd_0 = ssh_cmd[0].strip().lower()
            self.is_putty = ssh_cmd_0.endswith("plink") or ssh_cmd_0.endswith("plink.exe")
            self.is_paramiko = ssh_cmd_0 == "paramiko"
            full_ssh = ssh_cmd[:]
            full_ssh += get_ssh_args(params, ssh_cmd)
            if username:
                params["username"] = username
            if self.nostrict_host_check.get_active():
                full_ssh += ["-o", "StrictHostKeyChecking=no"]
            if params["type"] == MODE_NESTED_SSH:
                params["type"] = "ssh"
                params["proxy_host"] = self.config.proxy_host
                params["proxy_port"] = self.config.proxy_port
                params["proxy_username"] = self.config.proxy_username
                params["proxy_password"] = self.config.proxy_password
                full_ssh += get_ssh_proxy_args(params, ssh_cmd)
            params["host"] = host
            params["local"] = is_local(self.config.host)
            params["full_ssh"] = full_ssh
            params["password"] = password
            params["display_name"] = f"ssh://{self.config.host}:{self.config.port}"
        elif self.config.mode == "display":
            params["display"] = f":{self.config.port}"
            params["display_name"] = f":{self.config.port}"
        elif self.config.mode in ("socket", "unix-domain"):
            params["display"] = f":{self.config.port}"
            params["display_name"] = f"unix-domain:{self.config.port}"
        else:
            if self.config.mode not in (MODE_TCP, MODE_SSL, MODE_QUIC, MODE_WS, MODE_WSS):
                raise ValueError(f"invalid / unsupported mode {self.config.mode}")
            params["host"] = self.config.host
            params["local"] = is_local(self.config.host)
            params["port"] = int(self.config.port)
            params["display_name"] = f"{self.config.mode}://{self.config.host}:{self.config.port}"
            if self.config.mode in (MODE_WS, MODE_WSS):
                params["path"] = self.config.path
            # always try to populate ssl so we can auto-upgrade:
            try:
                params["ssl-options"] = get_ssl_options(params, self.config, [])
            except ImportError:
                # perhaps ssl is missing or blocked
                pass
            if self.config.mode in (MODE_SSL, MODE_WSS, MODE_QUIC) and self.nostrict_host_check.get_active():
                params["strict-host-check"] = False

        # print("connect_to(%s)" % params)
        # UGLY warning: the username may have been updated during display parsing,
        # or the config file may contain a username which is different from the default one
        # which is used for initializing the client during init,
        # so update the client now:
        configure_env(self.config.env)
        configure_logging(self.config, "attach")
        configure_network(self.config)
        self.start_client(params)

    def start_client(self, display_desc: dict) -> None:
        bypass_no_gtk()
        self.client = make_client(self.config)
        self.client.show_progress(30, "client configuration")
        self.client.init(self.config)
        self.client.show_progress(40, "loading user interface")
        self.client.init_ui(self.config)
        self.client.username = display_desc.get("username")

        def handshake_complete(*_args) -> None:
            self.client.show_progress(100, "Session connected")

        self.client.after_handshake(handshake_complete)
        self.set_info_text("Connecting...")
        start_thread(self.do_connect_builtin, "connect", daemon=True, args=(display_desc,))

    def ssh_failed(self, message) -> None:
        log(f"ssh_failed({message})")
        if not self.current_error:
            self.current_error = message
            self.set_info_text(message, True)

    def do_connect_builtin(self, display_desc: dict) -> None:
        log(f"do_connect_builtin({display_desc})")
        self.exit_code = None
        self.current_error = None
        self.set_info_text("Connecting.")
        self.set_sensitive(False)
        self.client.show_progress(60, "connecting")
        try:
            log("calling %s%s", connect_to,
                (display_desc, repr_ellipsized(str(self.config)), self.set_info_text, self.ssh_failed))
            conn = connect_to(display_desc, opts=self.config,
                              debug_cb=self.set_info_text, ssh_fail_cb=self.ssh_failed)
        except Exception as e:
            log("do_connect_builtin(%s) failed to connect", display_desc, exc_info=True)
            self.handle_exception(e)
            return
        log("connect_to(..)=%s, hiding launcher window, starting client", conn)
        GLib.idle_add(self.start_xpra_client, conn, display_desc)

    def start_xpra_client(self, conn, display_desc: dict) -> None:
        try:
            self.do_start_xpra_client(conn, display_desc)
        except Exception as e:
            log.error("Error: failed to start client", exc_info=True)
            self.handle_exception(e)

    def do_start_xpra_client(self, conn, display_desc: dict) -> None:
        log("do_start_XpraClient(%s, %s) client=%s", conn, display_desc, self.client)
        self.client.encoding = self.config.encoding
        self.client.display_desc = display_desc
        self.client.make_protocol(conn)
        self.set_info_text("Network connection established")
        log("start_XpraClient() client initialized")

        if self.config.password:
            self.client.password = self.config.password

        def do_quit(*args) -> None:
            log("do_quit%s", args)
            self.clean_client()
            self.destroy()
            Gtk.main_quit()

        def handle_client_quit(exit_launcher=False) -> None:
            w = self.window
            log("handle_quit(%s) window=%s", exit_launcher, w)
            self.clean_client()
            if exit_launcher:
                # give time for the main loop to run once after calling cleanup
                GLib.timeout_add(100, do_quit)
            else:
                if w:
                    self.set_sensitive(True)
                    GLib.idle_add(w.show)

        def reconnect(exit_code) -> bool:
            log("reconnect(%s) config reconnect=%s",
                exit_str(exit_code), self.config.reconnect)
            if not self.config.reconnect or exit_code not in RETRY_EXIT_CODES:
                return False
            self.clean_client()
            # give time for the main loop to run once after calling cleanup
            GLib.timeout_add(100, self.start_client, display_desc)
            return True

        def warn_and_quit_override(exit_code, warning) -> None:
            log("warn_and_quit_override(%s, %s)", exit_code, warning)
            password_warning = warning.find("invalid password") >= 0
            if password_warning:
                self.password_warning()
            elif reconnect(exit_code):
                return
            if self.exit_code is None:
                self.exit_code = exit_code
            err = exit_code != 0 or password_warning
            if not self.current_error:
                self.current_error = warning
                self.set_info_text(warning, err)
            handle_client_quit(not err)

        def quit_override(exit_code) -> None:
            log("quit_override(%s)", exit_code)
            if reconnect(exit_code):
                return
            if self.exit_code is None:
                self.exit_code = exit_code
                estr = exit_str(exit_code)
                self.set_info_text(estr, exit_code != ExitCode.OK)
            handle_client_quit(self.exit_code == 0)

        self.client.warn_and_quit = warn_and_quit_override
        self.client.quit = quit_override

        def after_handshake() -> None:
            self.set_info_text("Handshake complete")

        self.client.after_handshake(after_handshake)

        def first_ui_received(*_args) -> None:
            self.set_info_text("Running")
            self.window.hide()

        self.client.connect("first-ui-received", first_ui_received)
        if Gtk.main_level() > 0:
            # no need to start a new main loop:
            self.client.gtk_main = noop
        try:
            r = self.client.run()
            log("client.run() returned %s", r)
        except Exception as e:
            log.error("client error", exc_info=True)
            self.handle_exception(e)
        if self.client.gtk_main == noop:
            return
        log("exit_launcher=%s", self.exit_launcher)
        # if we're using "autoconnect",
        # the main loop was running from here,
        # so we have to force exit if the launcher window had been closed:
        if self.exit_launcher:
            sys.exit(0)

    def password_ok(self, *_args) -> None:
        with IgnoreWarningsContext():
            self.password_entry.modify_text(Gtk.StateType.NORMAL, black)

    def password_warning(self, *_args) -> None:
        with IgnoreWarningsContext():
            self.password_entry.modify_text(Gtk.StateType.NORMAL, red)
        self.password_entry.grab_focus()

    def update_options_from_gui(self) -> None:
        def pint(vstr, default=0) -> int:
            try:
                return int(vstr)
            except ValueError:
                return default

        self.config.host = self.host_entry.get_text()
        self.config.ssh_port = pint(self.ssh_port_entry.get_text())
        self.config.port = pint(self.port_entry.get_text(), -1)
        self.config.username = self.username_entry.get_text()
        self.config.password = self.password_entry.get_text()
        self.config.path = self.path_entry.get_text()
        self.config.autoconnect = self.autoconnect.get_active()

        self.config.proxy_host = self.proxy_host_entry.get_text()
        self.config.proxy_port = pint(self.proxy_port_entry.get_text())
        self.config.proxy_username = self.proxy_username_entry.get_text()
        self.config.proxy_password = self.proxy_password_entry.get_text()
        if self.is_putty:
            self.config.proxy_key = self.proxy_key_entry.get_text()

        mode_enc = self.mode_combo.get_active_text().lower()
        if mode_enc.startswith(MODE_TCP):
            self.config.mode = MODE_TCP
            if mode_enc.find("aes") > 0:
                self.config.encryption = "AES-" + mode_enc.split("aes-")[1].upper()
        elif mode_enc in (MODE_SSL, MODE_QUIC, MODE_SSH, MODE_WS, MODE_WSS, MODE_NESTED_SSH):
            self.config.mode = mode_enc
            self.config.encryption = ""
        log("update_options_from_gui() %s",
            (self.config.username, self.config.password, self.config.mode, self.config.encryption,
             self.config.host, self.config.port, self.config.ssh_port))

    def update_gui_from_config(self) -> None:
        mode = (self.config.mode or "").lower()
        active = 0
        for i, e in enumerate(get_connection_modes()):
            if e.lower() == mode:
                active = i
                break
        self.mode_combo.set_active(active)

        def get_port(vstr, default_port="") -> str:
            try:
                iport = int(vstr)
                if 0 <= iport < 2 ** 16:
                    return str(iport)
            except ValueError:
                pass
            return str(default_port)

        dport = DEFAULT_PORT
        if mode in (MODE_SSH, MODE_NESTED_SSH):
            # not required, so don't specify one
            dport = ""
        self.port_entry.set_text(get_port(self.config.port, dport))
        self.ssh_port_entry.set_text(get_port(self.config.ssh_port))
        # proxy bits:
        self.proxy_host_entry.set_text(self.config.proxy_host)
        self.proxy_port_entry.set_text(get_port(self.config.proxy_port))
        username = self.config.username
        proxy_username = self.config.proxy_username
        self.username_scb.set_active(username == proxy_username)
        self.proxy_username_entry.set_text(proxy_username)
        password = self.config.password
        proxy_password = self.config.proxy_password
        self.password_scb.set_active(password == proxy_password)
        self.proxy_password_entry.set_text(proxy_password)
        if self.is_putty:
            self.proxy_key_entry.set_text(self.config.proxy_key)
        self.username_entry.set_text(username)
        self.password_entry.set_text(password)
        self.host_entry.set_text(self.config.host)
        self.path_entry.set_text(self.config.path)

    def close_window(self, *_args) -> None:
        w = self.window
        if w:
            self.window = None
            w.destroy()

    def destroy(self, *args) -> bool:
        log("destroy%s", args)
        self.exit_launcher = True
        self.clean_client()
        self.close_window()
        Gtk.main_quit()
        return False

    def update_options_from_url(self, url: str) -> None:
        from xpra.scripts.parsing import parse_URL
        address, props = parse_URL(url)
        pa = address.split("://")
        if pa[0] in (MODE_TCP, MODE_SSH, MODE_SSL, MODE_QUIC, MODE_WS, MODE_WSS) and len(pa) >= 2:
            props["mode"] = pa[0]
            host = pa[1]
            if host.find("@") > 0:
                username, host = host.rsplit("@", 1)
                if username.find(":") > 0:
                    username, password = username.rsplit(":", 1)
                    props["password"] = password
                props["username"] = username
            if host.find(":") > 0:
                host, port = host.rsplit(":", 1)
                props["port"] = int(port)
            props["host"] = host
        self._apply_props(props)

    def update_options_from_file(self, filename: str) -> None:
        props = read_config(filename)
        log("update_options_from_file(%s) props=%s", filename, props)
        self._apply_props(props)

    def _apply_props(self, props: dict[str, Any]) -> None:
        # we rely on "ssh_port" being defined on the config object
        # so try to load it from file, and define it if not present:
        options = validate_config(props,
                                  extras_types=LAUNCHER_OPTION_TYPES,
                                  extras_validation=get_launcher_validation())
        for k, v in options.items():
            fn = k.replace("-", "_")
            setattr(self.config, fn, v)
        self.config_keys = self.config_keys.union(set(props.keys()))
        self.parse_ssh()
        log("_apply_props(%s) populated config with keys '%s', ssh=%s", props, options.keys(), self.config.ssh)

    def choose_session_file(self, title: str, action, action_button, callback: Callable):
        file_filter = Gtk.FileFilter()
        file_filter.set_name("Xpra")
        file_filter.add_pattern("*.xpra")
        choose_file(self.window, title, action, action_button, callback, file_filter)

    def save_clicked(self, *_args) -> None:
        self.update_options_from_gui()

        def do_save(filename) -> None:
            # make sure the file extension is .xpra
            if os.path.splitext(filename)[-1] != ".xpra":
                filename += ".xpra"
            save_config(filename, self.config, self.config_keys, extras_types=LAUNCHER_OPTION_TYPES)

        self.choose_session_file("Save session settings to file", Gtk.FileChooserAction.SAVE, Gtk.STOCK_SAVE, do_save)

    def load_clicked(self, *_args) -> None:
        def do_load(filename) -> None:
            self.update_options_from_file(filename)
            self.update_gui_from_config()

        self.choose_session_file("Load session settings from file", Gtk.FileChooserAction.OPEN, Gtk.STOCK_OPEN, do_load)


# on some platforms like win32, we don't have stdout
# and this is a GUI application, so show a dialog with the error instead
def exception_dialog(title: str) -> None:
    md = Gtk.MessageDialog(None, Gtk.DialogFlags.DESTROY_WITH_PARENT,
                           Gtk.MessageType.INFO, Gtk.ButtonsType.CLOSE, title)
    md.format_secondary_text(traceback.format_exc())
    md.show_all()

    def close_dialog(*_args) -> None:
        md.destroy()
        Gtk.main_quit()

    md.connect("response", close_dialog)
    md.connect("close", close_dialog)
    Gtk.main()


def main(argv) -> int:
    from xpra.platform import program_context
    from xpra.log import enable_color
    with program_context("Xpra-Launcher", "Xpra Connection Launcher"):
        enable_color()
        return do_main(argv)


def do_main(argv: list[str]) -> int:
    from xpra.util.system import SIGNAMES
    from xpra.scripts.main import InitExit, InitInfo
    from xpra.platform.gui import init as gui_init, ready as gui_ready
    from xpra.gtk.util import init_display_source

    init_display_source()

    gui_init()
    try:
        from xpra.scripts.parsing import parse_cmdline, fixup_debug_option
        options, args = parse_cmdline(argv)
        debug = fixup_debug_option(options.debug)
        if debug:
            enable_debug_for(*debug.split(","))
    except InitInfo as e:
        print(str(e))
        return 0
    except InitExit as e:
        return e.status
    except Exception:
        exception_dialog("Error parsing command line")
        return 1

    # allow config to be debugged:
    from xpra.scripts import config
    config.debug = log.debug

    try:
        app = ApplicationWindow()

        def handle_signal(signum) -> None:
            app.show()
            client = app.client
            if client:
                client.cleanup()
            else:
                Gtk.main_quit()
            GLib.timeout_add(1000, app.set_info_text, "got signal " + SIGNAMES.get(signum, signum))

        register_os_signals(handle_signal, "Client Launcher")
        has_file = len(args) == 1
        if has_file:
            app.update_options_from_file(args[0])
            # the compressors and packet encoders cannot be changed from the UI
            # so apply them now:
            configure_env(app.config.env)
            configure_network(app.config)
        debug = fixup_debug_option(app.config.debug)
        if debug:
            enable_debug_for(*debug.split(","))
        app.create_window_with_config()
    except Exception:
        exception_dialog("Error creating launcher form")
        return 1
    try:
        if app.config.autoconnect:
            # file says we should connect,
            # do that only (not showing UI unless something goes wrong):
            GLib.idle_add(app.do_connect)
        if not has_file:
            app.reset_errors()
        if not app.config.autoconnect or app.config.debug:
            if OSX:
                from xpra.platform.darwin.gui import wait_for_open_handlers, force_focus
                if has_file:
                    force_focus()
                    app.show()
                else:
                    def open_file(filename: str) -> None:
                        log("open_file(%s)", filename)
                        app.update_options_from_file(filename)
                        # the compressors and packet encoders cannot be changed from the UI
                        # so apply them now:
                        configure_network(app.config)
                        app.update_gui_from_config()
                        if app.config.autoconnect:
                            app.__osx_open_signal = True
                            GLib.idle_add(app.do_connect)
                        else:
                            force_focus()
                            app.show()

                    def open_URL(url: str) -> None:
                        log("open_URL(%s)", url)
                        app.__osx_open_signal = True
                        app.update_options_from_url(url)
                        # the compressors and packet encoders cannot be changed from the UI
                        # so apply them now:
                        configure_network(app.config)
                        app.update_gui_from_config()
                        GLib.idle_add(app.do_connect)

                    wait_for_open_handlers(app.show, open_file, open_URL)
            else:
                app.show()
        gui_ready()
        app.run()
    except KeyboardInterrupt:
        pass
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections import namedtuple
import dbus.service  # @UnresolvedImport
import dbus.types  # @UnresolvedImport

from xpra.dbus.helper import dbus_to_native, native_to_dbus
from xpra.dbus.common import init_session_bus
from xpra.server.dbus.server_base import DBUS_Server_Base, INTERFACE, BUS_NAME
from xpra.util.parsing import parse_scaling_value, from0to100
from xpra.common import ConnectionMessage
from xpra.log import (
    Logger,
    add_debug_category, remove_debug_category,
    disable_debug_for, enable_debug_for,
)

log = Logger("dbus", "server")

Rectangle = namedtuple("Rectangle", ("x", "y", "width", "height"))


def n(*args):
    return dbus_to_native(*args)


def ni(*args):
    return int(n(*args))


def ns(*args):
    return str(n(*args))


def nb(*args):
    return bool(n(*args))


class DBUS_Server(DBUS_Server_Base):

    def __init__(self, server=None, extra=""):
        bus = init_session_bus()
        name = BUS_NAME
        if extra:
            name += extra.replace(".", "_").replace(":", "_")
        super().__init__(bus, server, name)
        self._properties |= {
            "idle-timeout": ("idle_timeout", ni),
            "server-idle-timeout": ("server_idle_timeout", ni),
            "name": ("session_name", ns),
        }

    @dbus.service.method(INTERFACE, in_signature='i')
    def Focus(self, wid):
        wid = ni(wid)
        self.log(".Focus(%i)", wid)
        self.server.control_command_focus(wid)

    @dbus.service.method(INTERFACE, in_signature='')
    def Suspend(self):
        self.log(".Suspend()")
        self.server.control_command_suspend()

    @dbus.service.method(INTERFACE, in_signature='')
    def Resume(self):
        self.log(".Resume()")
        self.server.control_command_resume()

    @dbus.service.method(INTERFACE, in_signature='')
    def Ungrab(self):
        self.log(".Ungrab()")
        self.server.control_command_resume()

    @dbus.service.method(INTERFACE, in_signature='s')
    def Start(self, command):
        c = ns(command)
        self.log(".Start(%s)", c)
        self.server.do_control_command_start(True, c)

    @dbus.service.method(INTERFACE, in_signature='s')
    def StartChild(self, command):
        c = ns(command)
        self.log(".StartChild(%s)", c)
        self.server.do_control_command_start(False, c)

    @dbus.service.method(INTERFACE, in_signature='sb')
    def ToggleFeature(self, feature, state):
        f, s = ns(feature), ns(state)
        self.log(".ToggleFeature(%s, %s)", f, s)
        self.server.control_command_toggle_feature(f, s)

    @dbus.service.method(INTERFACE, in_signature='i')
    def KeyPress(self, keycode):
        k = ni(keycode)
        self.log(".KeyPress(%i)", k)
        self.server.control_command_key(str(k), press=True)

    @dbus.service.method(INTERFACE, in_signature='i')
    def KeyRelease(self, keycode):
        k = ni(keycode)
        self.log(".KeyRelease(%i)", k)
        self.server.control_command_key(str(k), press=False)

    @dbus.service.method(INTERFACE, in_signature='')
    def ClearKeysPressed(self):
        self.log(".ClearKeysPressed()")
        self.server.clear_keys_pressed()

    @dbus.service.method(INTERFACE, in_signature='ii')
    def SetKeyboardRepeat(self, repeat_delay, repeat_interval):
        d, i = ni(repeat_delay), ni(repeat_interval)
        self.log(".SetKeyboardRepeat(%i, %i)", d, i)
        self.server.set_keyboard_repeat(d, i)

    @dbus.service.method(INTERFACE, in_signature='iii')
    def MovePointer(self, wid, x, y):
        wid, x, y = ni(wid), ni(x), ni(y)
        self.log(".MovePointer(%i, %i, %i)", wid, x, y)
        device_id = -1
        self.server._move_pointer(device_id, ni(wid), (ni(x), ni(y)))

    @dbus.service.method(INTERFACE, in_signature='ib')
    def MouseClick(self, button, pressed):
        button, pressed = ni(button), nb(pressed)
        self.log(".MouseClick%s", (button, pressed))
        device_id = -1
        self.server.button_action(device_id, 0, None, button, pressed)

    @dbus.service.method(INTERFACE, in_signature='iiii')
    def SetWorkarea(self, x, y, w, h):
        x, y, w, h = ni(x), ni(y), ni(w), ni(h)
        self.log(".SetWorkarea%s", (x, y, w, h))
        workarea = Rectangle(x=x, y=y, width=w, height=h)
        self.server.set_workarea(workarea)

    @dbus.service.method(INTERFACE, in_signature='iiiii')
    def SetVideoRegion(self, wid, x, y, w, h):
        wid, x, y, w, h = ni(wid), ni(x), ni(y), ni(w), ni(h)
        self.log(".SetVideoRegion%s", (wid, x, y, w, h))
        self.server.control_command_video_region(wid, x, y, w, h)

    @dbus.service.method(INTERFACE, in_signature='ib')
    def SetVideoRegionEnabled(self, wid, enabled):
        wid, enabled = ni(wid), nb(enabled)
        self.log(".SetVideoRegionEnabled(%i, %s)", wid, enabled)
        self.server.control_command_video_region_enabled(wid, enabled)

    @dbus.service.method(INTERFACE, in_signature='ib')
    def SetVideoRegionDetection(self, wid, detection):
        wid, detection = ni(wid), nb(detection)
        self.log(".SetVideoRegionDetection(%i, %s)", wid, detection)
        self.server.control_command_video_region_detection(wid, detection)

    @dbus.service.method(INTERFACE, in_signature='iaai')
    def SetVideoRegionExclusionZones(self, wid, zones):
        wid = ni(wid)
        nzones = []
        for zone in zones:
            nzones.append([ni(x) for x in zone])
        log("SetVideoRegionExclusionZones(%i, %s)", wid, nzones)
        self.server.control_command_video_region_exclusion_zones(wid, nzones)

    @dbus.service.method(INTERFACE, in_signature='i')
    def ResetVideoRegion(self, wid):
        wid = ni(wid)
        self.log(".ResetVideoRegion(%i)", wid)
        self.server.control_command_reset_video_region(wid)

    @dbus.service.method(INTERFACE, in_signature='ii')
    def LockBatchDelay(self, wid, delay):
        wid, delay = ni(wid), ni(delay)
        self.log(".LockBatchDelay(%i, %i)", wid, delay)
        self.server.control_command_lock_batch_delay(wid, delay)

    @dbus.service.method(INTERFACE, in_signature='i')
    def UnlockBatchDelay(self, wid):
        wid = ni(wid)
        self.log(".UnlockBatchDelay(%i)", wid)
        self.server.control_command_unlock_batch_delay(wid)

    @dbus.service.method(INTERFACE, in_signature='', out_signature='a{is}')
    def ListWindows(self):
        d = {}
        for wid, window in self.server._id_to_window.items():
            try:
                d[wid] = window.get_property("title")
            except Exception:  # pragma: no cover
                d[wid] = str(window)
        self.log(".ListWindows()=%s", d)
        return d

    @dbus.service.method(INTERFACE, in_signature='s')
    def SetLock(self, lock):
        s = ns(lock)
        self.log(".SetLock(%s)", s)
        self.server.control_command_set_lock(s)

    @dbus.service.method(INTERFACE, in_signature='s')
    def SetSharing(self, sharing):
        s = ns(sharing)
        self.log(".SetSharing(%s)", s)
        self.server.control_command_set_sharing(s)

    @dbus.service.method(INTERFACE, in_signature='s')
    def SetUIDriver(self, uuid):
        s = ns(uuid)
        self.log(".SetUIDriver(%s)", s)
        self.server.control_command_set_ui_driver(s)

    @dbus.service.method(INTERFACE, in_signature='i')
    def SetIdleTimeout(self, value):
        nvalue = ni(value)
        self.log(".SetIdleTimeout(%s)", nvalue)
        self.server.control_command_idle_timeout(nvalue)

    @dbus.service.method(INTERFACE, in_signature='ii')
    def MoveWindowToWorkspace(self, wid, workspace):
        wid, workspace = ni(wid), ni(workspace)
        self.log(".MoveWindowToWorkspace(%i, %i)", wid, workspace)
        self.server.control_command_workspace(wid, workspace)

    @dbus.service.method(INTERFACE, in_signature='is')
    def SetWindowScaling(self, wid, scaling):
        wid, scaling = ni(wid), ns(scaling)
        self.log(".SetWindowScaling(%i, %s)", wid, scaling)
        s = parse_scaling_value(scaling)
        self.server.control_command_scaling(s, wid)

    @dbus.service.method(INTERFACE, in_signature='is')
    def SetWindowScalingControl(self, wid, scaling_control):
        wid = ni(wid)
        self.log(".SetWindowScalingControl(%i, %s)", wid, scaling_control)
        if scaling_control.lower() in ("auto", "on"):
            sc = None
        else:
            sc = from0to100(int(ns(scaling_control)))
        self.server.control_command_scaling_control(sc, wid)

    @dbus.service.method(INTERFACE, in_signature='is')
    def SetWindowEncoding(self, wid, encoding):
        wid, encoding = ni(wid), ns(encoding)
        self.log(".SetWindowEncoding(%i, %s)", wid, encoding)
        self.server.control_command_encoding(encoding, wid)

    @dbus.service.method(INTERFACE, in_signature='i')
    def RefreshWindow(self, wid):
        wid = ni(wid)
        self.log(".RefreshWindow(%i)", wid)
        self.server.control_command_refresh(wid)

    @dbus.service.method(INTERFACE, in_signature='ai')
    def RefreshWindows(self, window_ids):
        wids = [ni(x) for x in window_ids]
        self.log(".RefreshWindows(%s)", wids)
        self.server.control_command_refresh(*wids)

    @dbus.service.method(INTERFACE, in_signature='')
    def RefreshAllWindows(self):
        self.log(".RefreshAllWindows()")
        self.server.control_command_refresh(*self.server._id_to_window.keys())

    @dbus.service.method(INTERFACE, in_signature='')
    def ResetWindowFilters(self):
        self.log(".ResetWindowFilters()")
        self.server.reset_window_filters()

    @dbus.service.method(INTERFACE, in_signature='s')
    def EnableDebug(self, category):
        c = ns(category)
        self.log(".EnableDebug(%s)", c)
        add_debug_category(c)
        enable_debug_for(c)

    @dbus.service.method(INTERFACE, in_signature='s')
    def DisableDebug(self, category):
        c = ns(category)
        self.log(".DisableDebug(%s)", c)
        remove_debug_category(c)
        disable_debug_for(c)

    @dbus.service.method(INTERFACE, in_signature='isss')
    def SendNotification(self, nid, title, message, uuids):
        nid, title, message, uuids = ni(nid), ns(title), ns(message), ns(uuids)
        self.log(".SendNotification%s", (nid, title, message, uuids))
        self.server.control_command_send_notification(nid, title, message, uuids)

    @dbus.service.method(INTERFACE, in_signature='is')
    def CloseNotification(self, nid, uuids):
        nid, uuids = ni(nid), ns(uuids)
        self.log(".CloseNotification%s", (nid, uuids))
        self.server.control_command_close_notification(nid, uuids)

    @dbus.service.method(INTERFACE, in_signature='sii')
    def SetClipboardProperties(self, direction, max_copyin, max_copyout):
        # keep direction unchanged if not specified
        max_copyin, max_copyout = ni(max_copyin), ni(max_copyout)
        direction = ns(direction) or self.server.clipboard_direction
        self.log(".SetClipboardProperties%s", (direction, max_copyin, max_copyout))
        self.server.control_command_clipboard_direction(direction, max_copyin, max_copyout)

    @dbus.service.method(INTERFACE, in_signature='', out_signature='a{ss}')
    def ListClients(self):
        d = {}
        for p, source in self.server._server_sources.items():
            try:
                d[source.uuid] = str(p)
            except KeyError:
                d[str(source)] = str(p)
        self.log(".ListClients()=%s", d)
        return d

    @dbus.service.method(INTERFACE, in_signature='s')
    def DetachClient(self, uuid):
        s = ns(uuid)
        self.log(".DetachClient(%s)", s)
        for p, source in self.server._server_sources.items():
            if source.uuid == s:
                self.log("matched %s", source)
                self.server.disconnect_client(p, ConnectionMessage.DETACH_REQUEST)

    @dbus.service.method(INTERFACE)
    def DetachAllClients(self):
        self.log(".DetachAllClients() will detach: %s", self.server._server_sources)
        for p in self.server._server_sources.keys():
            self.server.disconnect_client(p, ConnectionMessage.DETACH_REQUEST)

    @dbus.service.method(INTERFACE, in_signature='as')
    def SendUIClientCommand(self, args):
        nargs = n(args)
        log("SendUIClientCommand(%s)", nargs)
        for src in self.server._server_sources.values():
            if src.ui_client:
                src.send_client_command(*nargs)

    @dbus.service.method(INTERFACE, in_signature='', out_signature='a{sv}', async_callbacks=("callback", "errback"))
    def GetAllInfo(self, callback, errback):
        self.log(".GetAllInfo()")

        def gotinfo(_proto, info):
            try:
                v = dbus.types.Dictionary((str(k), native_to_dbus(v)) for k, v in info.items())
                # v =  native_to_dbus(info)
                log("native_to_dbus(..)=%s", v)
                callback(v)
            except Exception as e:  # pragma: no cover
                log("GetAllInfo:gotinfo", exc_info=True)
                errback(str(e))

        v = self.server.get_all_info(gotinfo)
        self.log(".GetAllInfo()=%s", v)
        return v

    @dbus.service.method(INTERFACE, in_signature='s', out_signature='a{sv}', async_callbacks=("callback", "errback"))
    def GetInfo(self, subsystem, callback, errback):
        self.log(".GetInfo(%s)", subsystem)

        def gotinfo(_proto, info):
            sub = info.get(subsystem)
            try:
                v = dbus.types.Dictionary((str(k), native_to_dbus(v)) for k, v in sub.items())
                log("native_to_dbus(..)=%s", v)
                callback(v)
            except Exception as e:  # pragma: no cover
                log("GetInfo:gotinfo", exc_info=True)
                errback(str(e))

        v = self.server.get_all_info(gotinfo)
        self.log(".GetInfo(%s)=%s", subsystem, v)
        return v



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import signal

from xpra.exit_codes import ExitCode, ExitValue


def os_signal(sig, frame) -> None:
    sys.exit(ExitCode.IO_ERROR)


def main(args) -> ExitValue:
    from xpra.platform import program_context
    with program_context("Wait-for-X11", "Wait for X11"):
        signal.signal(signal.SIGINT, os_signal)
        signal.signal(signal.SIGTERM, os_signal)
        from xpra.x11.bindings.wait_for_x_server import wait_for_x_server
        if args:
            display = args[0]
        else:
            display = os.environ.get("DISPLAY", "")
        try:
            wait_for_x_server(display)
            return ExitCode.OK
        except SystemExit as e:  # NOSONAR @SuppressWarnings("python:S5727")
            return e.code
        except RuntimeError as e:
            sys.stderr.write(f"{e}\n")
            sys.stderr.flush()
            return ExitCode.FAILURE


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1:])
